import warnings
from enum import IntEnum

import numpy as np
import pandas as pd

try:
    import pyarrow  # pylint: disable=unused-import
except ImportError:
    PYARROW_PANDAS_BACKEND_AVAILABLE = False
else:
    PYARROW_PANDAS_BACKEND_AVAILABLE = pd.__version__ >= "2.0"


class DataType(IntEnum):
    """Class describe possible data types."""

    NUMERIC = 1
    CATEGORICAL = 2
    BOOLEAN = 3
    DATE = 4
    UNKNOWN = 5
    MISSING = 6
    UNIQUE = 7

    def __str__(self):
        return self.name.lower()


# pylint: disable=too-many-return-statements
def infer_data_type(series: pd.Series) -> DataType:
    """Infers the data type of the series passed in.

    Parameters
    ----------
    series : pd.Series
        Series from which to infer data type.

    Returns
    -------
    DataType
        Inferred custom edvart data type.
    """
    if series.empty:
        return DataType.UNKNOWN
    if is_missing(series):
        return DataType.MISSING
    if is_boolean(series):
        return DataType.BOOLEAN
    if is_date(series):
        return DataType.DATE
    if is_unique(series):
        return DataType.UNIQUE
    if is_categorical(series):
        return DataType.CATEGORICAL
    if is_numeric(series):
        return DataType.NUMERIC

    return DataType.UNKNOWN


def is_unique(series: pd.Series) -> bool:
    """Heuristic to tell if a series is categorical with only unique values.

    Parameters
    ----------
    series : pd.Series
        Series from which to infer data type.

    Returns
    -------
    bool
        Boolean indicating whether series contains only unique values.
    """
    return is_categorical(series) and series.nunique() == len(series)


def is_numeric(series: pd.Series) -> bool:
    """
    Heuristic to tell if a series contains numbers only.

    Parameters
    ----------
    series : pd.Series
        Series from which to infer data type.

    Returns
    -------
    bool
        Boolean indicating whether series contains only numbers.
    """
    if is_missing(series):
        return False
    return pd.api.types.is_numeric_dtype(series)


def is_missing(series: pd.Series) -> bool:
    """Function to tell if the series contains only missing values.

    Parameters
    ----------
    series : pd.Series
        Series from which to infer data type.

    Returns
    -------
    bool
        True if all values in the series are missing, False otherwise.
    """
    return series.isnull().all()


def is_categorical(series: pd.Series, unique_value_count_threshold: int = 10) -> bool:
    """Heuristic to tell if a series is categorical.

    Parameters
    ---
    series : pd.Series
        Series from which to infer data type.
    unique_value_count_threshold : int
        The number of unique values of the series has to be less than or equal to this number for
        the series to satisfy one of the requirements to be a categorical series.

    Returns
    ---
    bool
        Boolean indicating if series is categorical.
    """
    return (
        not is_missing(series)
        and not is_boolean(series)
        and not is_date(series)
        and (
            (
                series.nunique() <= unique_value_count_threshold
                and pd.api.types.is_integer_dtype(series)
            )
            or pd.api.types.is_string_dtype(series)
            or pd.api.types.is_object_dtype(series)
        )
    )


def is_boolean(series: pd.Series) -> bool:
    """Heuristic which tells if a series contains only boolean values.

    Parameters
    ----------
    series : pd.Series
        Series from which to infer data type.

    Returns
    -------
    bool
        Boolean indicating if series is boolean.
    """
    return not is_missing(series) and (
        pd.api.types.is_bool_dtype(series) or set(series.unique()) <= {1, 0, pd.NA}
    )


def is_date(series: pd.Series) -> bool:
    """Heuristic which tells if a series is of type date.

    Parameters
    ----------
    series : pd.Series
        Series from which to infer data type.

    Returns
    -------
    bool
        Boolean indicating if series is of type datetime.
    """
    if isinstance(series.dtype, pd.PeriodDtype):
        return True
    if is_missing(series) or is_numeric(series):
        return False
    contains_numerics = np.any(series.astype(str).str.isnumeric())
    if contains_numerics:
        return False
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            converted_series = pd.to_datetime(series.dropna(), errors="coerce")
    except ValueError:
        return False
    return converted_series.notna().all()



from typing import Any, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
from IPython.display import Markdown, display
from scipy import signal

from edvart.data_types import is_numeric
from edvart.decorators import check_index_time_ascending
from edvart.report_sections.code_string_formatting import get_code, total_dedent
from edvart.report_sections.section_base import Section, Verbosity


class ShortTimeFT(Section):
    """Generates Short-time discrete Fourier transform spectrogram plot subsection.

    Parameters
    ----------
    sampling_rate : int
        The time series will be considered as samples from a lower-frequency at this rate, i.e.
        frequencies in multiples of (1 / sampling rate) will be analyzed.
    window_size : int
        Size of window to perform DFT on to obtain Short-time Fourier transform.
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the generated code in the exported notebook.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    """

    def __init__(
        self,
        sampling_rate: int,
        window_size: int,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
    ):
        if sampling_rate <= 0:
            raise ValueError(f"sampling_rate has to be a positive integer, not {sampling_rate}")
        if window_size <= 0:
            raise ValueError(f"window_size has to a positive integer, not {window_size}")
        super().__init__(verbosity, columns)
        self.sampling_rate = sampling_rate
        self.window_size = window_size

    @property
    def name(self) -> str:
        return "Short-time Fourier Transform"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                total_dedent(
                    """
                    from edvart.report_sections.timeseries_analysis import ShortTimeFT
                    short_time_ft = ShortTimeFT.short_time_ft
                    """
                )
            ]
        return [
            "from IPython.display import display, Markdown",
            "from edvart.data_types import is_numeric",
            "import matplotlib.pyplot as plt",
            "from scipy import signal",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)
        default_call = (
            f"show_short_time_ft(df=df, sampling_rate={self.sampling_rate}"
            f", window_size={self.window_size}"
        )
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = get_code(show_short_time_ft) + "\n\n" + default_call

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates Short-time Fourier transform spectrogram in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        show_short_time_ft(
            df=df,
            sampling_rate=self.sampling_rate,
            window_size=self.window_size,
            columns=self.columns,
        )


@check_index_time_ascending
def show_short_time_ft(
    df: pd.DataFrame,
    sampling_rate: int,
    window_size: int,
    columns: Optional[List[str]] = None,
    overlap: Optional[int] = None,
    log: bool = True,
    window: Union[str, Tuple, "array-like"] = "hamming",
    scaling: str = "spectrum",
    figsize: Tuple[float, float] = (20, 7),
    colormap: Any = "viridis",
    freq_min: Optional[float] = None,
    freq_max: Optional[float] = None,
) -> None:
    """Generates Short-time discrete Fourier transform spectrogram plot.

    Parameters
    ----------
    df : pd.DataFrame
        Data to analyze.
    sampling_rate : int
        The time series will be considered as samples from a lower-frequency at this rate, i.e.
        frequencies in multiples of (1 / sampling rate) will be analyzed.
    window_size : int
        Size of window to perform DFT on to obtain Short-time Fourier transform.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    overlap : int, optional
        How many samples adjacent windows overlap by. Default `window_size // 8`.
    log : bool (default = True)
        Whether to color plot according by linear-scale amplitude or log-scale (in decibel).
    window : str (default = "hann")
        Type of weighting of individual samples in a window. If string or tuple, it is passed
        to `scipy.signal.get_window`. If array-like, each term is weight for the corresponding
        sample within the windows
    scaling : str (default = "density")
        Selects between computing the power spectral density ("density") with units of V**2/Hz
        and computing the power spectrum ("spectrum") with units of V**2,
        if input values are measured in V and sampling_rate is measured in Hz.
    figsize : Tuple[float, float] (default = (20, 7))
        Size of generated spectral plot figure.
    colormap : Any
        Any seaborn-compatible colormap.
    freq_min : float, optional
        Lowest frequency to show in the plot. All computed frequencies are shown by default.
    freq_max : float, optional
        Highest frequency to show in the plot. All computed frequencies are shown by default.

    Raises
    ------
    ValueError
        If the input data is not indexed by time in ascending order.
    """
    if sampling_rate <= 0:
        raise ValueError(f"Sampling rate has to be a positive integer, not {sampling_rate}")
    if window_size <= 0:
        raise ValueError(f"window_size has to a positive integer, not {window_size}")
    if columns is None:
        columns = [col for col in df.columns if is_numeric(df[col])]
    else:
        for col in columns:
            if not is_numeric(df[col]):
                raise ValueError(f"Cannot perform STFT for non-numeric column {col}")
    index_freq = pd.infer_freq(df.index) or ""
    for col in columns:
        display(Markdown(f"---\n### {col}"))
        freqs, times, sx = signal.spectrogram(  # pylint: disable=invalid-name
            # interpolate to get samples at regular time intervals
            df[col].interpolate(method="time"),
            fs=sampling_rate,
            window=window,
            nperseg=window_size,
            # Overlap defaults to window_size // 8
            noverlap=overlap,
            scaling=scaling,
        )

        # Add small positive value to avoid 0 in log
        y = 10 * np.log10(sx + 1e-12) if log else sx

        _fig, ax = plt.subplots(figsize=figsize)
        ax.pcolormesh(times, freqs, y, cmap=colormap)

        ax.set_ylabel(f"Frequency [1/({sampling_rate}{index_freq})]")
        ax.set_xlabel("Time")
        ax.set_ylim(freq_min, freq_max)
        # Show times from index in xticks
        ax.set_xticklabels(
            df.index[list(map(lambda time: int(time * sampling_rate), ax.get_xticks()[:-1]))]
        )
        plt.show()



from enum import IntEnum
from typing import Any, Dict, List, Optional

import nbformat.v4 as nbfv4
import pandas as pd
from IPython.display import Markdown, display

from edvart.decorators import check_index_time_ascending
from edvart.report_sections.section_base import ReportSection, Verbosity
from edvart.report_sections.timeseries_analysis import (
    Autocorrelation,
    BoxplotsOverTime,
    FourierTransform,
    RollingStatistics,
    SeasonalDecomposition,
    ShortTimeFT,
    StationarityTests,
    TimeSeriesLinePlot,
)


# pylint: disable=invalid-name
class TimeseriesAnalysisSubsection(IntEnum):
    """Enum of all implemented timeseries analysis subsections."""

    TimeSeriesLinePlot = 0
    RollingStatistics = 1
    BoxplotsOverTime = 2
    SeasonalDecomposition = 3
    StationarityTests = 4
    Autocorrelation = 5
    FourierTransform = 6
    ShortTimeFT = 7

    def __str__(self):
        return self.name


class TimeseriesAnalysis(ReportSection):
    """Generates the Timeseries analysis section of the report.

    Contains an enum TimeseriesAnalysisSubsection of possible subsections.

    Parameters
    ----------
    subsections : List[TimeseriesAnalysisSubsection], optional
        List of subsections to include.
        All subsection in TimeseriesAnalysisSubsection are included by default, except for
        FourierTransform, which is only included if `sampling_rate` is set and ShortTimeFT, which
        is only included if `sampling_rate` and `stft_window_size` are both set.
    verbosity : Verbosity
        Generated code verbosity global to the Overview sections.
        If subsection verbosities are None, then they will be overridden by this parameter.
    columns : List[str], optional
        Columns to include in timeseries analysis. Each column is treated as a separate time series.
        All columns are used by default.
    verbosity_series_line_plot : Verbosity, optional
        Time series line plot subsection code verbosity.
    verbosity_rolling_statistics: Verbosity, optional
        Rolling statistics interactive plot subsection code verbosity.
    verbosity_boxplots_over_time: Verbosity, optional
        Boxplots grouped over time intervals subsection code verbosity.
    verbosity_seasonal_decomposition: Verbosity, optional
        Seasonal decomposition subsection code verbosity.
    verbosity_stationarity_tests: Verbosity, optional
        Stationarity tests subsection code verbosity.
    verbosity_autocorrelation: Verbosity, optional
        Autocorrelation and partial autocorrelation plot subsection code verbosity.
    verbosity_fourier_transform: Verbosity, optional
        Discrete Fourier transform plot subsection code verbosity.
    verbosity_short_time_ft: Verbosity, optional
        Short-time discrete Fourier transform plot subsection code verbosity.
    sampling_rate: int, optional
        Sampling rate of the time-series, i.e., how many samples form one period. For example,
        if your time-series contains hourly data and want to investigate daily frequencies, use 24.
        If not set, Fourier transform and Short-time Fourier transform will not be included.
    stft_window_size: int
        Window size for Short-time Fourier transform, which will not be included if this parameter
        is not set.
    """

    def __init__(
        self,
        subsections: Optional[List[TimeseriesAnalysisSubsection]] = None,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        verbosity_time_series_line_plot: Optional[Verbosity] = None,
        verbosity_rolling_statistics: Optional[Verbosity] = None,
        verbosity_boxplots_over_time: Optional[Verbosity] = None,
        verbosity_seasonal_decomposition: Optional[Verbosity] = None,
        verbosity_stationarity_tests: Optional[Verbosity] = None,
        verbosity_autocorrelation: Optional[Verbosity] = None,
        verbosity_fourier_transform: Optional[Verbosity] = None,
        verbosity_short_time_ft: Optional[Verbosity] = None,
        sampling_rate: Optional[int] = None,
        stft_window_size: Optional[int] = None,
    ):
        self.sampling_rate = sampling_rate
        self.stft_window_size = stft_window_size

        verbosity_time_series_line_plot = verbosity_time_series_line_plot or verbosity
        verbosity_rolling_statistics = verbosity_rolling_statistics or verbosity
        verbosity_boxplots_over_time = verbosity_boxplots_over_time or verbosity
        verbosity_seasonal_decomposition = (
            verbosity_seasonal_decomposition
            if verbosity_seasonal_decomposition is not None
            else verbosity
        )
        verbosity_stationarity_tests = verbosity_stationarity_tests or verbosity
        verbosity_autocorrelation = verbosity_autocorrelation or verbosity
        verbosity_fourier_transform = verbosity_fourier_transform or verbosity
        verbosity_short_time_ft = verbosity_short_time_ft or verbosity

        self.subsection_verbosities = {
            TimeseriesAnalysisSubsection.TimeSeriesLinePlot: verbosity_time_series_line_plot,
            TimeseriesAnalysisSubsection.RollingStatistics: verbosity_rolling_statistics,
            TimeseriesAnalysisSubsection.BoxplotsOverTime: verbosity_boxplots_over_time,
            TimeseriesAnalysisSubsection.SeasonalDecomposition: verbosity_seasonal_decomposition,
            TimeseriesAnalysisSubsection.StationarityTests: verbosity_stationarity_tests,
            TimeseriesAnalysisSubsection.Autocorrelation: verbosity_autocorrelation,
            TimeseriesAnalysisSubsection.FourierTransform: verbosity_fourier_transform,
            TimeseriesAnalysisSubsection.ShortTimeFT: verbosity_short_time_ft,
        }

        enum_to_implementation = {
            TimeseriesAnalysisSubsection.TimeSeriesLinePlot: TimeSeriesLinePlot(
                verbosity_time_series_line_plot, columns
            ),
            TimeseriesAnalysisSubsection.RollingStatistics: RollingStatistics(
                verbosity_rolling_statistics, columns
            ),
            TimeseriesAnalysisSubsection.BoxplotsOverTime: BoxplotsOverTime(
                verbosity_boxplots_over_time, columns
            ),
            TimeseriesAnalysisSubsection.SeasonalDecomposition: SeasonalDecomposition(
                verbosity_seasonal_decomposition, columns
            ),
            TimeseriesAnalysisSubsection.StationarityTests: StationarityTests(
                verbosity_stationarity_tests, columns
            ),
            TimeseriesAnalysisSubsection.Autocorrelation: Autocorrelation(
                verbosity_autocorrelation, columns
            ),
        }
        # Add FT and STFT only if required parameters specified
        if sampling_rate is not None:
            enum_to_implementation[
                TimeseriesAnalysisSubsection.FourierTransform
            ] = FourierTransform(sampling_rate, verbosity_fourier_transform, columns)
            if stft_window_size is not None:
                enum_to_implementation[TimeseriesAnalysisSubsection.ShortTimeFT] = ShortTimeFT(
                    sampling_rate, stft_window_size, verbosity_short_time_ft, columns
                )
            elif (
                subsections is not None and TimeseriesAnalysisSubsection.ShortTimeFT in subsections
            ):
                raise ValueError(
                    "Need to set an `stft_window_size` to plot Short-time Fourier transform."
                )
        elif subsections is not None:
            if TimeseriesAnalysisSubsection.FourierTransform in subsections:
                raise ValueError("Need to set a `sampling_rate` to plot Fourier transform.")
            if TimeseriesAnalysisSubsection.ShortTimeFT in subsections:
                raise ValueError(
                    "Need to set a `sampling_rate` to plot Short-time Fourier transform."
                )

        self.default_subsections_to_show = list(enum_to_implementation.keys())

        if subsections is None:
            self.subsections_to_show = self.default_subsections_to_show
        else:
            self.subsections_to_show = subsections

        self.subsections_to_show_with_low_verbosity = [
            sub
            for sub in self.subsections_to_show
            if self.subsection_verbosities[sub] == Verbosity.LOW
        ]

        subsections_implementations = [
            enum_to_implementation[sub] for sub in self.subsections_to_show
        ]

        super().__init__(subsections_implementations, verbosity, columns)

    @property
    def name(self) -> str:
        return "Timeseries analysis"

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Add cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)

        if self.verbosity == Verbosity.LOW:
            code = "show_timeseries_analysis(df=df"
            if self.subsections_to_show_with_low_verbosity != self.default_subsections_to_show:
                arg_subsections_names = [
                    f"TimeseriesAnalysisSubsection.{str(sub)}"
                    for sub in self.subsections_to_show_with_low_verbosity
                ]
                code += f", subsections={arg_subsections_names}".replace("'", "")
            stft_included = (
                TimeseriesAnalysisSubsection.ShortTimeFT
                in self.subsections_to_show_with_low_verbosity
            )
            include_sampling_rate = self.sampling_rate is not None and (
                stft_included
                or TimeseriesAnalysisSubsection.FourierTransform
                in self.subsections_to_show_with_low_verbosity
            )
            if include_sampling_rate:
                code += f", sampling_rate={self.sampling_rate}"
                if self.stft_window_size is not None and stft_included:
                    code += f", stft_window_size={self.stft_window_size}"

            if self.columns is not None:
                code += f", columns={self.columns}"
            code += ")"
            cells.append(nbfv4.new_code_cell(code))

            for sub in self.subsections:
                if sub.verbosity > Verbosity.LOW:
                    sub.add_cells(cells=cells, df=df)
        else:
            super().add_cells(cells, df=df)

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"]
        """
        if self.verbosity == Verbosity.LOW:
            imports = {
                "from edvart.report_sections.timeseries_analysis.timeseries_analysis"
                " import show_timeseries_analysis"
            }
            if self.subsections_to_show_with_low_verbosity != self.default_subsections_to_show:
                imports.add(
                    "from edvart.report_sections.timeseries_analysis.timeseries_analysis"
                    " import TimeseriesAnalysisSubsection"
                )
            for sub in self.subsections:
                if sub.verbosity > Verbosity.LOW:
                    imports.update(sub.required_imports())

            return list(imports)
        return super().required_imports()

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        display(Markdown(self.get_title(section_level=1)))
        super().show(df)


@check_index_time_ascending
def show_timeseries_analysis(
    df: pd.DataFrame,
    subsections: Optional[List[TimeseriesAnalysisSubsection]] = None,
    columns: Optional[List[str]] = None,
    sampling_rate: Optional[int] = None,
    stft_window_size: Optional[int] = None,
) -> None:
    """Generate timeseries analysis for df.

    Parameters
    ----------
    df : pd.DataFrame
        Data to be analyzed.
    subsections : List[TimeseriesAnalysisSubsection], optional
        Subsections to include in the analysis. All subsections are included by default.
    columns : List[str], optional
        Subset of columns of df to consider in timeseries analysis.
        All columns are used by default.
    sampling_rate : int, optional
        Sampling rate of the time-series, i.e., how many samples form one period. For example,
        if your timeseries contains hourly data and you want to investigate daily frequencies,
        use 24.
        If not set, Fourier transform and Short-time Fourier transform will not be included.
    stft_window_size : int, optional
        Window size for Short-time Fourier transform. Short-time Fourier transform will not be
        included if this parameter is not set.

    Raises
    ------
    ValueError
        If the input data is not indexed by time in ascending order.
    """
    if columns is not None:
        df = df[columns]

    timeseries_analysis = TimeseriesAnalysis(
        subsections=subsections,
        verbosity=Verbosity.LOW,
        columns=columns,
        sampling_rate=sampling_rate,
        stft_window_size=stft_window_size,
    )

    for sub in timeseries_analysis.subsections:
        sub.show(df)



from typing import Any, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import pandas as pd
import statsmodels.api as sm
from IPython.display import Markdown, display

from edvart.data_types import is_numeric
from edvart.decorators import check_index_time_ascending
from edvart.report_sections.code_string_formatting import get_code
from edvart.report_sections.section_base import Section, Verbosity


class SeasonalDecomposition(Section):
    """Generates seasonal decomposition subsection.

    Each timeseries represented by one column is decomposed into
    trend, seasonal and residual (noise) components. This is a primitive decomposition.
    The seasonal component is first removed by applying a convolution filter to the data.
    The average of this smoothed series for each period is the returned seasonal component.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the generated code in the exported notebook.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    period : int, optional
        Period to use when modelling seasonal component. If None, period is inferred from
        frequency of df.index, provided `pd.infer_freq` is able to infer the frequency.
        Otherwise, this parameter has to be set manually.
    model : str (default = "additive")
        Can be either "multiplicative" or "additive".
        If "additive", series is modelled as series = trend + seasonal + noise
        If "multiplicative", series is modelled as series = trend * seasonal * noise
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        period: Optional[int] = None,
        model: str = "additive",
    ):
        self.model = model
        self.period = period
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Seasonal decomposition"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                "from edvart.report_sections.timeseries_analysis.seasonal_decomposition"
                " import show_seasonal_decomposition"
            ]
        return [
            "from IPython.display import display, Markdown",
            "import matplotlib.pyplot as plt",
            "import statsmodels.api as sm",
            "from edvart.data_types import is_numeric",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)
        default_call = "show_seasonal_decomposition(df=df"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        if self.period is not None:
            default_call += f", period={self.period}"
        default_call += f", model='{self.model}')"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = get_code(show_seasonal_decomposition) + "\n\n" + default_call

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates seasonal decomposition plot(s) in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        show_seasonal_decomposition(
            df=df, columns=self.columns, model=self.model, period=self.period
        )


@check_index_time_ascending
def show_seasonal_decomposition(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    period: Optional[int] = None,
    model: str = "additive",
    figsize: Tuple[float, float] = (20, 10),
) -> None:
    """Generate the seasonal decomposition plot.

    Parameters
    ----------
    df : pd.DataFrame
        Data to analyze.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    period : int, optional
        Period to use when modelling seasonal component. If None, period is inferred from
        frequency of df.index, provided `pd.infer_freq` is able to infer the frequency.
        Otherwise, this parameter has to be set manually.
    model : str (default = "additive")
        Can be either "multiplicative" or "additive".
        If "additive", series is modelled as series = trend + seasonal + noise
        If "multiplicative", series is modelled as series = trend * seasonal * noise
    figsize : Tuple[float, float] (default = (20, 10))
        Size of the whole figure for one column (i.e. includes plots of all components).

    Raises
    ------
    ValueError
        If the input data is not indexed by time in ascending order.
    """
    df = df.interpolate(method="time")
    if pd.infer_freq(df.index) is None and period is None:
        display(
            Markdown(
                "<div class='alert alert-block alert-warning'>"
                "Period could not be inferred, please set the `period` parameter"
                " to a suitable value. Not plotting seasonal decomposition."
                "</div>"
            )
        )
        return
    if columns is None:
        columns = [col for col in df.columns if is_numeric(df[col])]

    for col in columns:
        if not is_numeric(df[col]):
            raise ValueError(f"Cannot plot decomposition for non-numeric column {col}")
        display(Markdown(f"---\n### {col}"))
        decomposition = sm.tsa.seasonal_decompose(df[col], period=period, model=model)
        fig = decomposition.plot()
        fig.set_size_inches(*figsize)
        fig.axes[0].set_title(None)
        fig.axes[0].set_ylabel("Original")
        fig.axes[-1].set_ylabel("Residual")
        plt.show()



import itertools
from enum import IntEnum
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
import seaborn as sns
from IPython.display import Markdown, display

from edvart import utils
from edvart.data_types import is_boolean, is_categorical, is_numeric
from edvart.report_sections.code_string_formatting import get_code
from edvart.report_sections.section_base import ReportSection, Section, Verbosity


# pylint:disable=invalid-name
class BivariateAnalysisSubsection(IntEnum):
    """Enum of all implemented bivariate analysis subsections."""

    CorrelationPlot = 0
    PairPlot = 1
    ContingencyTable = 2

    def __str__(self):
        return self.name


class BivariateAnalysis(ReportSection):
    """Generates the Bivariate analysis section of the report.

    Contains an enum BivariateAnalysisSubsection of possible subsections.

    Parameters
    ----------
    subsections : List[BivariateAnalysisSubsection], optional
        List of subsections to include.
        All subsection in BivariateAnalysisSubsection are included by default.
    verbosity : Verbosity (default = Verbosity.LOW)
        Generated code verbosity global to the Bivariate analysis sections.
        If subsection verbosities are None, then they will be overridden by this parameter.
    columns : List[str], optional
        Columns on which to do bivariate analysis.
        If none of columns_x, columns_y and columns_pairs is specified bivariate analysis is
        performed on all pairs of `columns`.
        Ignored if `columns_x` and `columns_y` is specified. Ignored in contingency table
        if `columns_x` and `columns_y` or `columns_pairs` is specified.
        All columns are used by default.
    columns_x : List[str], optional
        If specified, correlations and pairplots are performed on the cartesian product of
        `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
    columns_y : List[str], optional
        If specified, correlations and pairplots are performed on the cartesian product of
        `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.
    columns_pairs : List[str], optional
        List of columns pairs on which to perform bivariate analysis.
        Used primarily in contingency tables.
        If specified, `columns`, `columns_x` and `columns_y` are ignored in contingency tables.
        Ignored in pairplots and correlations unless `columns_pairs` is specified and none of
        `columns`, `columns_x`, `columns_y` is specified. In that case, the first elements of each
        pair are treated as `columns_x` and the second elements as `columns_y` in pairplots and
        correlations.
    verbosity_correlations : Verbosity, optional
        Correlation plots subsection code verbosity.
    verbosity_pairplot: Verbosity, optional
        Pairplot subsection code verbosity.
    verbosity_contingency_table: Verbosity, optional
        Contingency table subsection code verbosity.
    color_col : str, optional
        Name of column according to use for coloring of the bivariate analysis subsections.
        Coloring is currently supported in pairplot.

    Raises
    ------
    ValueError
        If exactly one of `columns_x`, `columns_y` is specified.
    """

    # By default use all subsections
    _DEFAULT_SUBSECTIONS_TO_SHOW = list(BivariateAnalysisSubsection)

    def __init__(
        self,
        subsections: Optional[List[BivariateAnalysisSubsection]] = None,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        columns_pairs: Optional[List[Tuple[str, str]]] = None,
        verbosity_correlations: Optional[Verbosity] = None,
        verbosity_pairplot: Optional[Verbosity] = None,
        verbosity_contingency_table: Optional[Verbosity] = None,
        color_col: Optional[str] = None,
    ):
        verbosity_correlations = verbosity_correlations or verbosity
        verbosity_pairplot = verbosity_pairplot or verbosity
        verbosity_contingency_table = verbosity_contingency_table or verbosity

        # Store subsection verbosities
        self.subsection_verbosities = {
            BivariateAnalysisSubsection.CorrelationPlot: verbosity_correlations,
            BivariateAnalysisSubsection.PairPlot: verbosity_pairplot,
            BivariateAnalysisSubsection.ContingencyTable: verbosity_contingency_table,
        }

        if subsections is None:
            self.subsections_to_show = self._DEFAULT_SUBSECTIONS_TO_SHOW
        else:
            self.subsections_to_show = subsections
        self.subsections_to_show_with_low_verbosity = [
            sub
            for sub in self.subsections_to_show
            if self.subsection_verbosities[sub] == Verbosity.LOW
        ]

        if (columns_x is None) != (columns_y is None):
            raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
        # For analyses which do not take columns_pairs, prepare columns_x and columns_y in case
        # columns_pairs is the only parameter specified
        if columns is None and columns_x is None and columns_pairs is not None:
            columns_x_no_pairs = [pair[0] for pair in columns_pairs]
            columns_y_no_pairs = [pair[1] for pair in columns_pairs]
        else:
            columns_x_no_pairs = columns_x
            columns_y_no_pairs = columns_y
        enum_to_implementation = {
            BivariateAnalysisSubsection.CorrelationPlot: CorrelationPlot(
                verbosity_correlations, columns, columns_x_no_pairs, columns_y_no_pairs
            ),
            BivariateAnalysisSubsection.PairPlot: PairPlot(
                verbosity_pairplot,
                columns,
                columns_x_no_pairs,
                columns_y_no_pairs,
                color_col=color_col,
            ),
            BivariateAnalysisSubsection.ContingencyTable: ContingencyTable(
                verbosity_contingency_table, columns, columns_x, columns_y, columns_pairs
            ),
        }

        subsections_implementations = [
            enum_to_implementation[sub] for sub in self.subsections_to_show
        ]
        super().__init__(subsections_implementations, verbosity, columns)

        self.columns_x = columns_x
        self.columns_y = columns_y
        self.columns_pairs = columns_pairs
        self.color_col = color_col

    @property
    def name(self) -> str:
        return "Bivariate analysis"

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)
        if self.verbosity == Verbosity.LOW:
            code = "show_bivariate_analysis(df=df"
            if self.subsections_to_show_with_low_verbosity != self._DEFAULT_SUBSECTIONS_TO_SHOW:
                arg_subsections_names = [
                    f"BivariateAnalysisSubsection.{str(sub)}"
                    for sub in self.subsections_to_show_with_low_verbosity
                ]

                code += f", subsections={arg_subsections_names}".replace("'", "")
            if self.columns_x is not None:
                code += f", columns_x={self.columns_x}"
                code += f", columns_y={self.columns_y}"
            elif self.columns is not None:
                code += f", columns={self.columns}"
            if self.columns_pairs is not None:
                code += f", columns_pairs={self.columns_pairs}"
            if self.color_col is not None:
                code += f", color_col='{self.color_col}'"
            code += ")"
            cells.append(nbfv4.new_code_cell(code))
            for sub in self.subsections:
                if sub.verbosity > Verbosity.LOW:
                    sub.add_cells(cells=cells, df=df)
        else:
            super().add_cells(cells=cells, df=df)

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """
        if self.verbosity != Verbosity.LOW:
            return super().required_imports()

        imports = {"from edvart.report_sections.bivariate_analysis import show_bivariate_analysis"}
        if self.subsections_to_show_with_low_verbosity != self._DEFAULT_SUBSECTIONS_TO_SHOW:
            imports.add(
                "from edvart.report_sections.bivariate_analysis import BivariateAnalysisSubsection"
            )
        for subsec in self.subsections:
            if subsec.verbosity > Verbosity.LOW:
                imports.update(subsec.required_imports())

        return list(imports)

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        display(Markdown(self.get_title(section_level=1)))
        super().show(df)


def show_bivariate_analysis(
    df: pd.DataFrame,
    subsections: Optional[List[BivariateAnalysisSubsection]] = None,
    columns: Optional[List[str]] = None,
    columns_x: Optional[List[str]] = None,
    columns_y: Optional[List[str]] = None,
    columns_pairs: Optional[List[Tuple[str, str]]] = None,
    color_col: Optional[str] = None,
) -> None:
    """Generates bivariate analysis for df.

    Parameters
    ----------
    df : pd.DataFrame
        Data to be analyzed
    subsections : List[BivariateAnalysisSubsection], optional
        Subsections to include in the analysis. All subsections are included by default.
    columns : List[str], optional
        Columns on which to do bivariate analysis.
        If none of columns_x, columns_y and columns_pairs is specified bivariate analysis is
        performed on all pairs of `columns`.
        Ignored if `columns_x` and `columns_y` is specified. Ignored in contingency table
        if `columns_x` and `columns_y` or `columns_pairs` is specified.
        All columns are used by default.
    columns_x : List[str], optional
        If specified, correlations and pairplots are performed on the cartesian product of
        `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
    columns_y : List[str], optional
        If specified, correlations and pairplots are performed on the cartesian product of
        `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.
    columns_pairs : List[str], optional
        List of columns pairs on which to perform bivariate analysis.
        Used primarily in contingency tables.
        If specified, `columns`, `columns_x` and `columns_y` are ignored in contingency tables.
        Ignored in pairplots and correlations unless `columns_pairs` is specified and none of
        `columns`, `columns_x`, `columns_y` is specified. In that case, the first elements of
        each pair are treated as `columns_x` and the second elements as `columns_y` in pairplots
        and correlations.
    color_col : str, optional
        Name of column to use for coloring of the bivariate analysis subsections.
        Coloring is currently supported in pairplot.
    """
    bivariate_analysis = BivariateAnalysis(
        subsections=subsections,
        verbosity=Verbosity.LOW,
        columns=columns,
        columns_x=columns_x,
        columns_y=columns_y,
        columns_pairs=columns_pairs,
        color_col=color_col,
    )

    for sub in bivariate_analysis.subsections:
        sub.show(df)


class CorrelationPlot(Section):
    """Generates the Correlation plot subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns on which to plot pair-wise correlation plot.
        If `columns_x` and `columns_y` are unspecified, then analysis is performed on all pairs
        of `columns`. Otherwise ignored.
        All columns are used by default.
    columns_x : List[str], optional
        If specified, correlation is plotted on the cartesian product
        of `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
    columns_y : List[str], optional
        If specified, correlation is plotted on the cartesian product
        of `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.

    Raises
    ------
    ValueError
        If exactly one of `columns_x`, `columns_y` is specified.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
    ):
        if (columns_x is None) != (columns_y is None):
            raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
        self.columns_x = columns_x
        self.columns_y = columns_y
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Correlation Plot"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.bivariate_analysis import plot_correlations"]
        return [
            "import numpy as np",
            "import seaborn as sns",
            "from IPython.display import display, Markdown",
            "import matplotlib.pyplot as plt",
            "%matplotlib inline",
            "from edvart import utils",
            "from edvart.data_types import is_numeric",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "plot_correlations(df=df"
        if self.columns_x is not None:
            default_call += f", columns_x={self.columns_x}"
            default_call += f", columns_y={self.columns_y}"
        elif self.columns is not None:
            default_call += f", columns={self.columns}"

        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(default_correlations)
                + "\n\n"
                + get_code(_get_columns_x_y)
                + "\n\n"
                + get_code(plot_correlation)
                + "\n\n"
                + get_code(plot_correlations)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates correlation plots in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        plot_correlations(
            df=df, columns=self.columns, columns_x=self.columns_x, columns_y=self.columns_y
        )


def default_correlations() -> Dict[str, Callable[[pd.DataFrame], pd.DataFrame]]:
    """
    Default dictionary of bivariate statistics that will be calculated for numerical columns.

    Returns
    -------
    dict
        A dictionary assigning correlation functions to correlation names.
        Dictionary signature: 'CorrelationName': corr_func
    """
    return {
        "pearson": utils.pearson,
        "spearman": utils.spearman,
        "kendall": utils.kendall,
    }


def _get_columns_x_y(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    columns_x: Optional[List[str]] = None,
    columns_y: Optional[List[str]] = None,
) -> Tuple[List[str], List[str]]:
    """
    Helper function to get columns_x and columns_y from provided parameters and dataset.

    Parameters
    ----------
    df : pd.DataFrame
        Dataset to analyze.
    columns : List[str], optional
        Specified columns parameter.
    columns_x : List[str], optional
        Specified columns_x parameter.
    columns_y : List[str], optional
        Specified columns_x parameter.


    Returns
    -------
    columns_x, columns_y: Tuple[List[str], List[str]]
        Resolved columns_x and columns_y.
    """
    if (columns_x is None) != (columns_y is None):
        raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
    if columns_x is None:
        if columns is None:
            columns = list(df.columns)
        columns_x = columns_y = columns
    columns_x = [col for col in columns_x if is_numeric(df[col])]
    columns_y = [col for col in columns_y if is_numeric(df[col])]

    return columns_x, columns_y


def plot_correlations(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    columns_x: Optional[List[str]] = None,
    columns_y: Optional[List[str]] = None,
    pearson: bool = True,
    spearman: bool = True,
    kendall: bool = True,
) -> None:
    """Plots multiple correlations.

    Parameters
    ----------
    df : pd.DataFrame
        Data based on which to plot correlations.
    columns : Optional[List[str]]
        List of columns of df to analyze. All numeric columns of df are used by default.
    pearson : bool (default = True)
        If True, Pearson correlation will be plotted.
    spearman : bool (default = True)
        If True, Spearman correlation will be plotted.
    kendall : bool (default = True)
        If True, Kendall correlation will be plotted.
    """
    correlations = default_correlations()

    # remove correlations which are not to be plotted
    if not pearson:
        correlations.pop("pearson", None)
    if not spearman:
        correlations.pop("spearman", None)
    if not kendall:
        correlations.pop("kendall", None)

    for corr_name, corr_func in correlations.items():
        plot_correlation(df, corr_name, corr_func, columns, columns_x, columns_y)


def plot_correlation(
    df: pd.DataFrame,
    corr_name: str,
    corr_func: Callable[[pd.DataFrame], pd.DataFrame],
    columns: Optional[List[str]] = None,
    columns_x: Optional[List[str]] = None,
    columns_y: Optional[List[str]] = None,
    size_factor: float = 0.7,
    font_size: float = 15,
    color_map: Optional[Any] = None,
) -> None:
    """Plots a correlation heatmap.

    Parameters
    ----------
    df : pd.DataFrame
        Data based on which to plot correlation.
    corr_name : str
        Name of correlation being plotted.
    corr_func : Callable[[np.ndarray, Optional[List[Tuple[str, str]]]], pd.DataFrame]
        Correlation function to be used.
    columns : Optional[List[str]]
        List of columns of df to analyze. All numeric columns of df are used by default.
    size_factor: float (default = 0.7)
        Size of each cell in the table.
    font_size : float (default = 15)
        Size of axis labels of the correlation plot.
    color_map : Any, optional
        Color map compatible with matplotlib/seaborn to use for the correlation plot.
        A divergent red-blue color map is used by default.

    Raises
    ------
    ValueError
        If exactly one of `columns_x`, `columns_y` is specified.
    """
    columns_x, columns_y = _get_columns_x_y(df, columns, columns_x, columns_y)
    if color_map is None:
        color_map = sns.diverging_palette(10, 220, sep=80, n=100)

    # show header
    display(Markdown(f"### {corr_name.capitalize()} Correlation"))

    # calculate correlation between columns
    corr = corr_func(df)
    corr = corr.loc[columns_x, columns_y]

    # plot correlation heatmap
    ax = sns.heatmap(
        corr, cmap=color_map, vmin=-1, vmax=1, square=True, xticklabels=True, yticklabels=True
    )

    # set axes
    ax.set_xticklabels(ax.get_xticklabels(), fontsize=font_size, rotation=90)
    ax.set_yticklabels(ax.get_yticklabels(), fontsize=font_size, rotation=0)
    ax.figure.set_size_inches(len(columns_x) * size_factor + 1, len(columns_y) * size_factor + 1)

    plt.show()


class PairPlot(Section):
    """Generates the Pairplot subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns on which to plot the pairplot.
        If `columns_x` and `columns_y` are unspecified, then analysis is performed on all pairs
        of `columns`. Otherwise ignored.
        All columns are used by default.
    columns_x : List[str], optional
        If specified, correlation is plotted
        on the cartesian product of `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
    columns_y : List[str], optional
        If specified, correlation is plotted
        on the cartesian product of `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.
    color_col : str, optional
        Name of column according to use for coloring of points and histogram in the pairplot.

    Raises
    ------
    ValueError
        If exactly one of `columns_x`, `columns_y` is specified.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        color_col: Optional[str] = None,
    ):
        if (columns_x is None) != (columns_y is None):
            raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
        self.columns_x = columns_x
        self.columns_y = columns_y
        self.color_col = color_col
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Pairplot"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.bivariate_analysis import plot_pairplot"]
        return [
            "from edvart.data_types import is_categorical, is_boolean",
            "import matplotlib.pyplot as plt",
            "import seaborn as sns",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "plot_pairplot(df=df"
        if self.columns_x is not None:
            default_call += f", columns_x={self.columns_x}"
            default_call += f", columns_y={self.columns_y}"
        elif self.columns is not None:
            default_call += f", columns={self.columns})"
        if self.color_col is not None:
            default_call += f", color_col='{self.color_col}'"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = get_code(plot_pairplot) + "\n\n" + default_call

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates pairplot in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        plot_pairplot(
            df=df,
            columns=self.columns,
            columns_x=self.columns_x,
            columns_y=self.columns_y,
            color_col=self.color_col,
        )


def plot_pairplot(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    columns_x: Optional[List[str]] = None,
    columns_y: Optional[List[str]] = None,
    allow_categorical: bool = False,
    color_col: Optional[str] = None,
) -> None:
    """Plot a pairplot for each pair of columns.

    Parameters
    ----------
    df : pd.DataFrame
        Data frame for which to plot pairplot.
    columns : Union[List[Tuple[str, str]] or List[str]], optional
        Which columns to plot pairplot for.
        All columns that are not categorical and are not boolean are used by default.
    columns_x : List[str], optional
        If specified, correlation is plotted
        on the cartesian product of `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
    columns_y : List[str], optional
        If specified, correlation is plotted
        on the cartesian product of `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.
    allow_categorical : bool (default = False)
        Whether to allow plotting of categorical columns. If False (default), then even
        explicitly specified columns will be excluded. If True, categorical columns are still
        excluded by default, unless explicitly specified via columns/columns_x/columns_y.
    color_col : str, optional
        Name of column according to use for coloring of points and histogram in the pairplot.

    Raises
    ------
    ValueError
        If exactly one of `columns_x`, `columns_y` is specified.
    """

    def include_column(col: str) -> bool:
        return not is_categorical(df[col]) and not is_boolean(df[col])

    if (columns_x is None) != (columns_y is None):
        raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
    if columns_x is None:
        if columns is None:
            columns = list(filter(include_column, df.columns))
        columns_x = columns
        columns_y = columns
    if not allow_categorical:
        columns_x = list(filter(include_column, columns_x))
        columns_y = list(filter(include_column, columns_y))
    sns.pairplot(df, x_vars=columns_x, y_vars=columns_y, hue=color_col)
    plt.tight_layout()
    plt.show()


class ContingencyTable(Section):
    """Generates the pairwise contingency tables subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns on which to show contingency tables.
        If `columns_x` and `columns_y` and `columns_pairs` are unspecified,
        then analysis performed on all pairs of `columns`. Otherwise ignored.
        Columns which contain only null values are always excluded. All columns are used by default.
    columns_x : List[str], optional
        If specified, contingency tables are plotted for each pair
        in the cartesian product of `columns_x` and `columns_y`.
        If `columns_x` is specified, then `columns_y` must also be specified.
        Columns which contain only null values are always excluded.
        Ignored if `columns_pairs` is specified.
    columns_y : List[str], optional
        If specified, contingency tables are plotted for each pair
        in the cartesian product of `columns_x` and `columns_y`.
        If `columns_y` is specified, then `columns_x` must also be specified.
        Columns which contain only null values are always excluded.
        Ignored if `columns_pairs` is specified.
    columns_pairs : List[Tuple[str, str]], optional
        If specified, contingency tables are plotted for exactly the specified pairs.
        Columns which contain only null values are always excluded, i.e. if at least one of the
        columns in a pair is excluded, then is excluded.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        columns_x: Optional[List[str]] = None,
        columns_y: Optional[List[str]] = None,
        columns_pairs: Optional[List[Tuple[str, str]]] = None,
    ):
        super().__init__(verbosity, columns)
        self.columns_x = columns_x
        self.columns_y = columns_y
        self.columns_pairs = columns_pairs

    @property
    def name(self) -> str:
        return "Contingency table"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.bivariate_analysis import contingency_tables"]
        return [
            "import itertools",
            "from edvart import utils",
            "import seaborn as sns",
            "import matplotlib.pyplot as plt",
            "from typing import Literal",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "contingency_tables(df=df"
        if self.columns_pairs is not None:
            default_call += f", columns_pairs={self.columns_pairs}"
        elif self.columns_x is not None:
            default_call += f", columns_x={self.columns_x}"
            default_call += f", columns_y={self.columns_y}"
        elif self.columns is not None:
            default_call += f", columns={self.columns}"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(contingency_tables)
                + "\n\n"
                + get_code(contingency_table)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates contingency table in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        contingency_tables(
            df=df,
            columns=self.columns,
            columns_x=self.columns_x,
            columns_y=self.columns_y,
            columns_pairs=self.columns_pairs,
        )


def contingency_tables(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    columns_x: Optional[List[str]] = None,
    columns_y: Optional[List[str]] = None,
    columns_pairs: Optional[List[Tuple[str, str]]] = None,
    table_threshold: int = 30,
) -> None:
    """Display a contingency table for each pairs of columns.

    Parameters
    ----------
    df : pd.DataFrame
        Data based on which to create a contingency table.
    columns : List[str], optional
        Which columns to generate pair-wise contingency tables for.
        Columns with more than `table_threshold` unique values are excluded.
        Columns which contain only null values are excluded.
        To override the excluded columns, specify `columns_pairs`.
        Ignored if `columns_x` and `columns_y` or `columns_pairs` is specified.
    columns_x : List[str], optional
        If specified, contingency tables are plotted for each pair
        in the cartesian product of `columns_x` and `columns_y`.
        Columns with more than `table_threshold` unique values are excluded.
        Columns which contain only null values are excluded.
        If `columns_x` is specified, then `columns_y` must also be specified.
        Ignored if `columns_pairs` is specified.
    columns_y : List[str], optional
        If specified, contingency tables are plotted for each pair
        in the cartesian product of `columns_x` and `columns_y`.
        Columns with more than `table_threshold` unique values are excluded.
        Columns which contain only null values are excluded.
        If `columns_y` is specified, then `columns_x` must also be specified.
        Ignored if `columns_pairs` is specified.
    columns_pairs : List[Tuple[str, str]], optional
        If specified, contingency tables are plotted for exactly the specified pairs.
    table_threshold : int (default = 30)
        Maximum number of unique values for a column to be used in contingency table.
        If non-positive, no columns are excluded according to this criterion.
    """

    def include_column(col: str) -> bool:
        n_unique = df[col].nunique()
        return (
            (table_threshold < 0 or n_unique <= table_threshold)
            and (not df[col].isnull().all())
            and n_unique > 1
        )

    if (columns_x is None) != (columns_y is None):
        raise ValueError("Either both or neither of columns_x, columns_y must be specified.")
    if columns is None:
        columns = list(df.columns)
    columns = [col for col in df.columns if include_column(col)]

    if columns_pairs is None:
        if columns_x is None:
            columns_pairs = list(itertools.combinations(columns, 2))
        else:
            columns_pairs = [
                (col_x, col_y)
                for (col_x, col_y) in itertools.product(columns_x, columns_y)
                # Filter out pairs of columns which contain the same column since
                # they make no sense in a contingency table
                if col_x != col_y and include_column(col_x) and include_column(col_y)
            ]

    for column1, column2 in columns_pairs:
        contingency_table(df, column1, column2)


def contingency_table(
    df: pd.DataFrame,
    columns1: Union[str, List[str]],
    columns2: Union[str, List[str]],
    include_total: bool = True,
    hide_zeros: bool = True,
    scaling_func: Callable[[np.ndarray], np.ndarray] = np.cbrt,
    colormap: Any = "Blues",
    size_factor: Union[float, Literal["auto"]] = "auto",
    fontsize: float = 15,
) -> None:
    """
    Show a colored contingency table for the two specified columns or lists of columns.

    Parameters
    ----------
    df : pd.DataFrame
        Data to analyze.
    columns1 : Union[str, List[str]]
        Name of column or list of column names to use in the vertical axis of the table.
    columns2 : Union[str, List[str]]
        Name of column or list of column names to use in the horizontal axis of the table.
    include_total : bool (default = True)
        Whether to include marginal counts.
    hide_zeros : bool (default = True)
        Whether to hide zero values in the table for better readability.
    scaling_func : Callable[["array-like"], "array-like"]
        Function to scale the values for the purpose of coloring for smaller spread.
        Cube root is used by default.
    colormap : Any (default = "Blues")
        Colormap compatible with matplotlib/seaborn.
    size_factor : float or "auto"
        Size of each cell in the table.
        If "auto", the cell size is automatically adjusted so that the numbers fit in the cells.
    fontsize : float (default = 15)
        Size of the font for axis labels.
    """
    if isinstance(columns1, str):
        columns1 = [columns1]
    if isinstance(columns2, str):
        columns2 = [columns2]
    table = pd.crosstab(
        [df[col] for col in columns1],
        [df[col] for col in columns2],
        margins_name="Total",
        margins=include_total,
    )

    annot = table.replace(0, "") if hide_zeros else table

    ax = sns.heatmap(
        scaling_func(table),
        annot=annot,
        fmt="",
        cbar=False,
        cmap=colormap,
        linewidths=0.1,
        xticklabels=1,
        yticklabels=1,
        annot_kws={"fontsize": fontsize},
        square=True,
    )
    if size_factor == "auto":
        n_digits_max = 1 + np.floor(np.log10(table.max().max()))
        size_factor = (
            # Constants chosen empirically to make the numbers fit in the cells
            0.18
            * max(4, n_digits_max)
        )
    ax.figure.set_size_inches(size_factor * len(table.columns), size_factor * len(table))
    # Set y axis
    ax.set_ylabel(ax.get_ylabel(), fontsize=fontsize)
    ax.set_yticklabels(ax.get_yticklabels(), rotation=0, fontsize=fontsize)

    # Set x axis
    xticklabels = ax.get_xticklabels()
    x_label_rotation = (
        0 if max(len(xticklabel.get_text()) for xticklabel in xticklabels) < 5 else 30
    )  # rotate if any x tick label is longer than 5 characters
    ax.set_xticklabels(xticklabels, fontsize=fontsize, rotation=x_label_rotation)

    ax.xaxis.tick_top()
    ax.set_xlabel(ax.get_xlabel(), fontsize=fontsize)
    ax.xaxis.set_label_position("top")

    # Visually separate the margins
    if include_total:
        ax.vlines(len(table.columns) - 1, ymin=0, ymax=len(table), color="grey")
        ax.hlines(len(table) - 1, xmin=0, xmax=len(table.columns), color="grey")

    plt.show()



import functools

import pandas as pd


def check_index_time_ascending(func):
    """
    Check whether the index of the DataFrame is sorted in ascending order.

    The DataFrame which is checked needs to be the first argument of the decorated function.

    Raises
    ------
    ValueError
        If the index is not a datetime or is not ascending.
    """

    @functools.wraps(func)
    def wrapper_check_index_ascending(df: pd.DataFrame, *args, **kwargs):
        if not (
            df.index.inferred_type.startswith("datetime") or df.index.inferred_type == "period"
        ):
            raise ValueError(
                "The index of the provided DataFrame is not a time index. Please reindex the data. "
                "See `edvart.utils.reindex_to_period` or `edvart.utils.reindex_to_datetime` "
            )
        if not df.index.is_monotonic_increasing:
            raise ValueError(
                "The index of the provided DataFrame is not in an ascending order. "
                "Please sort the index."
            )

        return func(df, *args, **kwargs)

    return wrapper_check_index_ascending



"""EDVART package."""

import logging
from importlib.metadata import PackageNotFoundError, version

from edvart import example_datasets
from edvart.report import DefaultReport, DefaultTimeseriesReport, Report, TimeseriesReport
from edvart.report_sections.dataset_overview import Overview
from edvart.report_sections.section_base import Verbosity

logging.basicConfig(level=logging.INFO)

try:
    __version__ = version("edvart")
except PackageNotFoundError:
    __version__ = "dev"



import warnings
from typing import Any, Dict, List, Optional

import nbformat.v4 as nbfv4
import pandas as pd
import plotly.graph_objects as go
from IPython.display import Markdown, display

from edvart.data_types import is_numeric
from edvart.decorators import check_index_time_ascending
from edvart.report_sections.code_string_formatting import get_code
from edvart.report_sections.section_base import Section, Verbosity


class RollingStatistics(Section):
    """Generates the rolling statistics interactive plot subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the generated code in the exported notebook.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    window_size : int (default = 20)
        Size of the rolling window to use when computing rolling statistics.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        window_size: int = 20,
    ):
        self.window_size = window_size
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Rolling statistics"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return [
                "from edvart.report_sections.timeseries_analysis.rolling_statistics"
                " import show_rolling_statistics"
            ]
        return [
            "import warnings",
            "from IPython.display import display, Markdown",
            "import plotly.graph_objects as go",
            "from edvart.data_types import is_numeric",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = "show_rolling_statistics(df=df"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = get_code(show_rolling_statistics) + "\n\n" + default_call

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates rolling statistics interactive plot(s) in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        show_rolling_statistics(df=df, columns=self.columns, window_size=self.window_size)


# pylint: disable=too-many-locals
@check_index_time_ascending
def show_rolling_statistics(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    window_size: int = 20,
    show_bands: bool = True,
    band_width: float = 1.0,
    show_std_dev: bool = True,
    color_mean: str = "#2040FF",
    color_band: str = "#90E0FF",
    color_std: str = "#CD5C5C",
) -> None:
    """Display rolling statistics interactive plot.

    Displays a separate plot for each column of df.

    Parameters
    ----------
    df : pd.DataFrame
        Data to analyze.
    columns : List[str], optional
        Columns to analyze. Only numeric columns can be analyzed.
        All numeric columns are used by default.
    window_size : int (default = 20)
        Size of the rolling window to use when computing rolling statistics.
    show_bands : bool (default = True)
        Whether to show lines delimiting the range
        [rolling_mean - band_width * rolling_std, rolling_mean + band_width * rolling_std]
    band_width : float (default = 1.)
        Multiple of standard deviation from mean to show bands at.
        Ignored if not showing bands.
    show_std_dev : bool (default = True)
        Whether to plot rolling standard deviation.
    color_mean : str (default = "#2040FF")
        Color of the line showing rolling mean.
    color_band : str (default = "#90E0FF")
        Color of the lines showing bands around rolling mean. Ignored if not showing bands.
    color_std : str (default = "#CD5C5C")
        Color of the line showing standard deviation. Ignored if not showing standard deviation.

    Raises
    ------
    ValueError
        If the input data is not indexed by time in ascending order.
    """
    if window_size >= len(df):
        window_size = len(df) - 1
        warnings.warn(f"Reducing window size to {window_size} due to insufficient samples.")
    if columns is None:
        columns = [col for col in df.columns if is_numeric(df[col])]
    else:
        for col in columns:
            if not is_numeric(df[col]):
                raise ValueError(f"Cannot plot rolling statistics for non-numeric column `{col}`")

    df_rolling = df[columns].rolling(window_size)
    df_rolling_mean = df_rolling.mean()[window_size - 1 :]
    df_rolling_std = df_rolling.std()[window_size - 1 :]
    index = df.index[window_size - 1 :]

    layout = dict(xaxis_rangeslider_visible=True)

    data = []
    for col in columns:
        data.append([])
        if show_std_dev:
            trace_std = go.Scatter(
                x=index,
                y=df_rolling_std[col],
                mode="lines",
                name="Rolling std. dev.",
                line={"color": color_std},
            )
            data[-1].append(trace_std)

        trace_mean = go.Scatter(
            x=index,
            y=df_rolling_mean[col],
            mode="lines",
            name="Rolling mean",
            line={"color": color_mean},
        )
        data[-1].append(trace_mean)

        if show_bands:
            # Plot upper band
            trace_mean_plus_std = go.Scatter(
                x=index,
                y=df_rolling_mean[col] + band_width * df_rolling_std[col],
                mode="lines",
                # pylint: disable=consider-using-f-string
                name=(
                    "Rolling mean + {} rolling std. dev.".format(
                        "" if band_width == 1 else str(band_width) + " * "
                    )
                ),
                line={"color": color_band},
            )
            # Plot lower band
            trace_mean_minus_std = go.Scatter(
                x=index,
                y=df_rolling_mean[col] - band_width * df_rolling_std[col],
                mode="lines",
                # pylint: disable=consider-using-f-string
                name=(
                    "Rolling mean - {} rolling std. dev.".format(
                        "" if band_width == 1 else str(band_width) + " * "
                    )
                ),
                line={"color": color_band},
            )
            data[-1].extend([trace_mean_plus_std, trace_mean_minus_std])

    for col_name, col_data in zip(columns, data):
        display(Markdown(f"---\n### {col_name}"))
        go.Figure(data=col_data, layout=layout).show()



from typing import Any, Callable, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
import seaborn as sns
from IPython.display import HTML, Markdown, display

from edvart import utils
from edvart.data_types import DataType, infer_data_type
from edvart.pandas_formatting import add_html_heading, dict_to_html, format_number, subcells_html
from edvart.report_sections.code_string_formatting import code_dedent, get_code
from edvart.report_sections.section_base import Section, Verbosity


class UnivariateAnalysis(Section):
    """Generates univariate analysis section of the report.

    Parameters
    -----------
    verbosity : Verbosity
        The verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns for which to do univariate analysis,
        all columns are used by default
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
    ):
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Univariate Analysis"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook.
        """
        if self.verbosity == Verbosity.LOW:
            return [
                "from edvart.report_sections.univariate_analysis import show_univariate_analysis"
            ]
        if self.verbosity == Verbosity.MEDIUM:
            return [
                code_dedent(
                    """from edvart.report_sections.univariate_analysis import (
                        bar_plot,
                        histogram,
                        numeric_statistics,
                        top_most_frequent,
                    )"""
                ),
                "from edvart import utils",
            ]
        # Verbosity.HIGH
        return [
            "import numpy as np",
            "from edvart import utils",
            "from IPython.display import display",
            "from IPython.display import HTML",
            "import matplotlib.pyplot as plt",
            "%matplotlib inline",
            "import seaborn as sns",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds univariate analysis cells to the list of cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        if self.columns is not None:
            df = df[self.columns]

        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=1))
        cells.append(section_header)

        if self.verbosity == Verbosity.HIGH:
            # Add cell with default stats dictionaries
            default_stats_dicts = nbfv4.new_code_cell(
                "# Default statistics dictionaries"
                + 2 * "\n"
                + get_code(default_descriptive_statistics)
                + 2 * "\n"
                + get_code(default_quantile_statistics)
            )
            cells.append(default_stats_dicts)
            # Add cell with function definitions
            func_defs = nbfv4.new_code_cell(
                get_code(top_most_frequent)
                + 2 * "\n"
                + get_code(bar_plot)
                + 2 * "\n"
                + (get_code(numeric_statistics))
                + 2 * "\n"
                + get_code(histogram)
            )
            cells.append(func_defs)
            # Add cell with dataframe rendering
            frame_rendering = nbfv4.new_code_cell(
                get_code(format_number)
                + 2 * "\n"
                + get_code(dict_to_html)
                + 2 * "\n"
                + get_code(add_html_heading)
                + 2 * "\n"
                + get_code(subcells_html)
            )
            cells.append(frame_rendering)

        if self.verbosity == Verbosity.LOW:
            if self.columns is None:
                code = "show_univariate_analysis(df=df)"
            else:
                code = f"show_univariate_analysis(df=df, columns={self.columns})"
            code_cell = nbfv4.new_code_cell(code)
            cells.append(code_cell)
        else:
            for col in df.columns:
                data_type = infer_data_type(df[col])
                data_type_name = str(data_type)
                column_header = nbfv4.new_markdown_cell(f"## *{col} - {data_type_name}*")
                cells.append(column_header)
                if data_type in (DataType.CATEGORICAL, DataType.BOOLEAN):
                    cell = nbfv4.new_code_cell(
                        code_dedent(
                            f"""
                            top_most_frequent(df['{col}'])
                            bar_plot(df['{col}'])"""
                        )
                    )
                elif data_type == DataType.UNIQUE:
                    cell = nbfv4.new_markdown_cell("Each value in the column is unique.")
                elif data_type == DataType.MISSING:
                    cell = nbfv4.new_markdown_cell("The column contains only missing values.")
                else:
                    if self.verbosity == Verbosity.MEDIUM:
                        cell = nbfv4.new_code_cell(
                            code_dedent(
                                f"""
                                numeric_statistics(df['{col}'])
                                histogram(df['{col}'])"""
                            )
                        )
                    else:
                        cell = nbfv4.new_code_cell(
                            code_dedent(
                                f"""
                                numeric_statistics(
                                    df['{col}'],
                                    descriptive_stats=default_descriptive_statistics(),
                                    quantile_stats=default_quantile_statistics()
                                )
                                histogram(df['{col}'])"""
                            )
                        )
                cells.append(cell)

    def show(self, df: pd.DataFrame) -> None:
        """Generates univariate analysis cell output in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        if self.columns is not None:
            df = df[self.columns]

        display(Markdown(self.get_title(section_level=1)))
        for col in df.columns:
            data_type = infer_data_type(df[col])
            data_type_name = str(data_type)
            display(Markdown(f"## *{col} - {data_type_name}*"))
            if data_type in (DataType.CATEGORICAL, DataType.BOOLEAN):
                top_most_frequent(df[col])
                bar_plot(df[col])
            elif data_type == DataType.UNIQUE:
                display(Markdown("Each value in the column is unique."))
            elif data_type == DataType.MISSING:
                display(Markdown("The column contains only missing values."))
            else:
                numeric_statistics(df[col])
                histogram(df[col])
            display(Markdown("---"))


def default_descriptive_statistics():
    """
    Return default descriptive statistics.

    Returns
    -------
    dict
        Dictionary with keys as statistics names and values with functions.
    """
    return {
        "Number of unique values": utils.num_unique_values,
        "Sum": utils.sum_,
        "Mean": utils.mean,
        "Mode": utils.mode,
        "Standard deviation": utils.std,
        "Mean absolute deviation": utils.mad,
        "Median absolute deviation": utils.median_absolute_deviation,
        "Coefficient of variation": utils.coefficient_of_variation,
        "Kurtosis": utils.kurtosis,
        "Skewness": utils.skewness,
    }


def default_quantile_statistics():
    """
    Return default quantile statistics.

    Returns
    -------
    dict
        Dictionary with keys as statistics names and values with functions.
    """
    return {
        "Minimum": utils.minimum,
        "Maximum": utils.maximum,
        "Q1": utils.quartile1,
        "Median": utils.median,
        "Q3": utils.quartile3,
        "Range": utils.value_range,
        "IQR": utils.iqr,
    }


def histogram(
    series: pd.Series,
    bins: Optional[int] = None,
    density: bool = False,
    box_plot: bool = True,
    figsize: Tuple[float, float] = (20, 7),
    distplot_kwargs: Optional[Dict[str, Any]] = None,
    boxplot_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    """Visualizes distribution of series with a histogram.

    Parameters
    ----------
    series : pd.Series
        Numerical series.
    bins : int, optional
        Number of bins of the histogram. If None, the number of bins is inferred.
    density : bool (default = False)
        If True, the area of the histogram bars will sum up to 1.
    box_plot : bool (default = True)
        If True, a horizontal box plot will be added above the histogram to visualize quartiles.
    figsize : Tuple[float, float] (default = (20, 7))
        Size of the figure of the visualization.
    distplot_kwargs : Dict[str, Any], optional
        Additional keyword arguments passed to seaborn.distplot
    boxplot_kwargs : Dict[str, Any], optional
        Additional keyword arguments passed to seaborn.boxplot
    """
    if distplot_kwargs is None:
        distplot_kwargs = {}
    if boxplot_kwargs is None:
        boxplot_kwargs = {}

    series = series.dropna()
    if bins is None:
        bin_edges = np.histogram_bin_edges(series, bins="auto")
        # Prevent too many bins, which slows down the visualization
        # "auto" uses the maximum of the Sturges and Freedman-Diaconis estimators
        # The Freedman-Diaconis rule can infer a huge number of bins for long-tailed distributions
        # There should never be a good reason to use more than 1000 bins
        if len(bin_edges) > 1000:
            bins = "sturges"
        else:
            bins = bin_edges

    if box_plot:
        _fig, (ax_box, ax_hist) = plt.subplots(
            nrows=2,
            sharex=True,
            gridspec_kw={"height_ratios": (0.15, 0.85)},
            figsize=figsize,
        )
        sns.boxplot(x=series, ax=ax_box, **boxplot_kwargs)
        sns.histplot(
            data=series,
            bins=bins,
            stat="density" if density else "count",
            ax=ax_hist,
            kde=False,
            **distplot_kwargs,
        )
        ax_box.set(xlabel="")
    else:
        plt.figure(figsize=figsize)
        sns.histplot(
            data=series,
            bins=bins,
            stat="density" if density else "count",
            kde=False,
            **distplot_kwargs,
        )
    plt.show()


def bar_plot(
    series: pd.Series,
    relative_count: bool = False,
    figsize: Tuple[float, float] = (20, 7),
    plotting_threshold: int = 50,
    **bar_plot_args: Dict[str, Any],
) -> None:
    """Plots a bar plot visualizing frequencies of series elements.

    Parameters
    ----------
    series : pd.Series
        Categorical series.
    relative_count : bool
        If True, the frequencies will be normalized by the series length.
    figsize : Tuple[float, float]
        Size of the bar plot.
    plotting_threshold : int
        If the number of unique values in the series is greater than this, no plot is created
        instead a warning is issued.
    bar_plot_args : Dict[str, Any]
        Additional kwargs passed to pandas.Series.bar.
    """
    if series.nunique() > plotting_threshold:
        display(
            Markdown(
                f"Number of unique values is greater than {plotting_threshold},"
                " not plotting bar plot."
            )
        )
    else:
        value_counts = series.value_counts()
        value_counts = value_counts / len(series) if relative_count else value_counts
        value_counts.plot.bar(figsize=figsize, **bar_plot_args)
        plt.show()


def numeric_statistics(
    series: pd.Series,
    descriptive_stats: Optional[Dict[str, Callable]] = None,
    quantile_stats: Optional[Dict[str, Callable]] = None,
    thousand_separator: str = " ",
) -> None:
    """Generates tables with statistics for numeric series.

    Parameters
    ----------
    series : pd.Series
        Numeric series.
    descriptive_stats : Dict[str, Callable], optional
        Descriptive statistics computed for the series.
        Dictionary format: {'statistic name': stat_func(pd.Series) -> float}
        If None, UnivariateAnalysis.default_descriptive_statistics is used.
    quantile_stats : Dict[str, Callable], optional
        Quantile statistics computed for the series.
        Dictionary format: {'statistic name': stat_func(pd.Series) -> float}
        If None, UnivariateAnalysis.default_quantile_statistics is used.
    thousand_separator : str
        Thousand separator for numbers in the tables, by default space.
    """
    if descriptive_stats is None:
        descriptive_stats = default_descriptive_statistics()
    if quantile_stats is None:
        quantile_stats = default_quantile_statistics()

    desc_stats = {
        stat_name: format_number(number=stat_func(series), thousand_separator=thousand_separator)
        for stat_name, stat_func in descriptive_stats.items()
    }
    quant_stats = {
        stat_name: format_number(number=stat_func(series), thousand_separator=thousand_separator)
        for stat_name, stat_func in quantile_stats.items()
    }
    # Render statistics tables side by side
    desc_stats_html = add_html_heading(dict_to_html(desc_stats), "Descriptive Statistics", 3)
    quant_stats_html = add_html_heading(dict_to_html(quant_stats), "Quantile Statistics", 3)
    html_table = [[desc_stats_html, quant_stats_html]]
    # Render
    display(HTML(subcells_html(html_table)))


def top_most_frequent(series: pd.Series, n_top: int = 5) -> None:
    """Generates a table with top n most frequent values in series.

    Parameters
    ----------
    series : pd.Series
        Categorical series
    n_top : int
        The number of most frequent values to include in the table.
    """
    frequent_values = dict_to_html(utils.top_frequent_values(series, n_top=n_top))
    frequent_values_html = add_html_heading(frequent_values, "Most frequent values", 3)
    display(HTML(frequent_values_html))


def show_univariate_analysis(df: pd.DataFrame, columns: Optional[List[str]] = None) -> None:
    """Generates univariate analysis for df.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to be analyzed
    columns : List[str], optional
        Subset of df columns to analyze, by default all columns of df are used
    """
    if columns is not None:
        df = df[columns]

    for col in df.columns:
        data_type = infer_data_type(df[col])
        data_type_name = str(data_type)
        display(Markdown(f"## *{col} - {data_type_name}*"))
        if data_type in (DataType.CATEGORICAL, DataType.BOOLEAN):
            top_most_frequent(df[col])
            bar_plot(df[col])
        elif data_type == DataType.UNIQUE:
            display(Markdown("Each value in the column is unique."))
        elif data_type == DataType.MISSING:
            display(Markdown("The column contains only missing values."))
        else:
            numeric_statistics(df[col])
            histogram(df[col])



import warnings
from typing import Any, Dict, List, Optional, Tuple

import nbformat.v4 as nbfv4
import pandas as pd
from IPython.display import Markdown, display

from edvart.plots import scatter_plot_2d
from edvart.report_sections.code_string_formatting import code_dedent, get_code
from edvart.report_sections.section_base import Section, Verbosity
from edvart.utils import select_numeric_columns

try:
    with warnings.catch_warnings():
        # Ignore warning from umap 0.5.0 that ParametricUMAP is unavailable
        # due to missing TensorFlow
        warnings.simplefilter("ignore", UserWarning)
        import umap
except ImportError as exc:
    raise ImportError("To use UMAP, install edvart with optional dependency 'umap'.") from exc


class UMAP(Section):
    """Plot a 2-dimensional UMAP embedding scatter plot.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        Columns to use in computing in the UMAP embedding. Only numeric columns can be used.
        All numeric columns are used by default.
    color_col : str, optional
        Name of column to color points on the plot by. Can be both numeric and categorical.
        By default, all points have the same color.
    interactive : bool (default = True)
        Whether to plot an interactive plot. The interactive plot also shows labels for each
        point on hover.
    n_neighbors : int (default = 15)
        UMAP embedding parameter controlling the balance between focusing on the
        local structure vs global structure. A low value means more focus on local structure.
    min_dist : int (default = 0.1)
        UMAP embedding parameter controlling how tightly points in the embedding are clumped
        together. A low value results in tighter clumping, which can show clusters or other
        similar structures, while a higher value encourages preservation of topological
        structure present in the input data.
    metric : str (default = "euclidean")
        UMAP embedding parameter controlling how distance is computed in the ambient space of
        the input data. Many different metrics are available, see UMAP documentation for a
        complete list.
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        color_col: Optional[str] = None,
        interactive: bool = True,
        n_neighbors: int = 15,
        min_dist: float = 0.1,
        metric: str = "euclidean",
    ):
        self.color_col = color_col
        self.interactive = interactive
        self.n_neighbors = n_neighbors
        self.min_dist = min_dist
        self.metric = metric
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "UMAP"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            return ["from edvart.report_sections.umap import plot_umap"]
        return [
            "from edvart.plots import scatter_plot_2d",
            "from edvart import utils",
            "import umap",
            "from edvart.data_types import is_numeric",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells. Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)

        default_call = code_dedent(
            f"""
            plot_umap(
                df=df,
                n_neighbors={self.n_neighbors},
                min_dist={self.min_dist},
                metric='{self.metric}',
            """.rstrip()
        )
        if self.columns is not None:
            default_call += f"\n    columns={self.columns},"
        if self.color_col is not None:
            default_call += f"\n    color_col='{self.color_col}',"
        if not self.interactive:
            default_call += "\n    interactive=False,"
        default_call += "\n)"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            code = (
                get_code(select_numeric_columns)
                + "\n\n"
                + get_code(plot_umap)
                + "\n\n"
                + default_call
            )

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates the UMAP plot section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        plot_umap(
            df=df,
            columns=self.columns,
            color_col=self.color_col,
            interactive=self.interactive,
            n_neighbors=self.n_neighbors,
            min_dist=self.min_dist,
            metric=self.metric,
        )


def plot_umap(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    color_col: Optional[str] = None,
    interactive: bool = True,
    n_neighbors: int = 15,
    min_dist: float = 0.1,
    metric: str = "euclidean",
    random_state: int = 42,
    figsize: Tuple[float, float] = (12, 12),
    opacity: float = 0.8,
    show_message: bool = True,
) -> None:
    """Plot a 2-dimensional UMAP embedding scatter plot.

    Parameters
    ----------
    df : pd.DataFrame
        Data to analyze.
    columns : List[str], optional
        Columns to use in computing in the UMAP embedding. Only numeric columns can be used.
        All numeric columns are used by default.
    color_col : str, optional
        Name of column to color points on the plot by. Can be both numeric and categorical.
        By default, all points have the same color.
    interactive : bool (default = True)
        Whether to plot an interactive plot. The interactive plot also shows labels for each
        point on hover.
    n_neighbors : int (default = 15)
        UMAP embedding parameter controlling the balance between focusing on the
        local structure vs global structure. A low value means more focus on local structure.
    min_dist : int (default = 0.1)
        UMAP embedding parameter controlling how tightly points in the embedding are clumped
        together. A low value results in tighter clumping, which can show clusters or other
        similar structures, while a higher value encourages preservation of topological
        structure present in the input data.
    metric : str (default = "euclidean")
        UMAP embedding parameter controlling how distance is computed in the ambient space of
        the input data. Many different metrics are available, see UMAP documentation for a
        complete list.
    random_state : int (default = 42)
        Random state for reproducibility of results, since UMAP is stochastic. If None, a random
        seed is used.
    figsize : Tuple[float, float] (default = (12, 12))
        Size of the resulting plot in inches.
    opacity : float (default = 0.8)
        Opacity of the points drawn in the scatter plot.
    show_message : bool (default = True)
        Whether to show a message informing the user to tune the embedding parameters.
    """
    columns = select_numeric_columns(df, columns)
    df = df.dropna(subset=columns)
    embedder = umap.UMAP(
        n_neighbors=n_neighbors, min_dist=min_dist, metric=metric, random_state=random_state
    )
    embedded = embedder.fit_transform(df[columns])

    scatter_plot_2d(
        df=df,
        x=embedded[:, 0],
        y=embedded[:, 1],
        color_col=color_col,
        interactive=interactive,
        figsize=figsize,
        opacity=opacity,
    )

    if show_message:
        print("UMAP requires proper setting of hyperparameters. ")
        print(
            "If results are unsatisfactory, consider trying different values of parameters"
            " `n_neighbors`, `min_dist` and `metric`."
        )



from typing import Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objs as go

from edvart.data_types import is_boolean, is_categorical, is_numeric

# Multiplier which makes plotly interactive plots (size in pixels) and
# matplotlib plots (size in inches) about the same size
_INCHES_TO_PIXELS = 64


def _is_color_col_categorical(column: pd.Series) -> bool:
    return is_categorical(column) or is_boolean(column) or not is_numeric(column)


# pylint: disable=too-many-locals, too-many-branches
def scatter_plot_2d(
    df: pd.DataFrame,
    x: Union[str, pd.Series, np.ndarray],
    y: Union[str, pd.Series, np.ndarray],
    color_col: Optional[str] = None,
    interactive: bool = True,
    figsize: Tuple[float, float] = (12, 12),
    opacity: float = 0.8,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    show_xticks: bool = False,
    show_yticks: bool = False,
    show_zerolines: bool = False,
    equal_scale_axes: bool = False,
) -> None:
    """Display a 2D scatter plot of x and y, with optional coloring of points by values in a column.

    Parameters
    ----------
    df : pd.DataFrame
        Data to plot.
    x : Union[str, pd.Series, np.ndarray]
        Name of column in df or flat array or series of x coordinates of plotted points.
    y : Union[str, pd.Series, np.ndarray]
        Name of column in df or flat array or series of y coordinates of plotted points.
    color_col : str, optional
        Name of column in df to color points in the plot by. Can be both numeric or categorical.
        By default, all points have the same color.
    interactive : bool (default = True)
        Whether to plot an interactive plot. The interactive plot also shows labels for each
        point on hover.
    figsize : Tuple[float, float] (default = (12, 12))
        Size of the resulting plot in inches.
    opacity : float (default = 0.8)
        Opacity of the points drawn in the scatter plot.
    xlabel: str, optional
        Label for the x axis. No label is displayed by default.
    ylabel: str, optional
        Label for the y axis. No label is displayed by default.
    show_xticks : bool (default = False)
        Whether to display ticks on the x axis.
    show_yticks : bool (default = False)
        Whether to display ticks on the y axis.
    show_zerolines : bool (default = False)
        Whether to display zero lines.
    equal_scale_axes : bool (default = False)
        Whether to make the x and y axes have the same scale.
    """
    if isinstance(x, str):
        x = df[x]
    if isinstance(y, str):
        y = df[y]
    plotting_func = _scatter_plot_2d_interactive if interactive else _scatter_plot_2d_noninteractive
    plotting_func(
        df=df,
        x=x,
        y=y,
        color_col=color_col,
        figsize=figsize,
        opacity=opacity,
        xlabel=xlabel,
        ylabel=ylabel,
        show_xticks=show_xticks,
        show_yticks=show_yticks,
        show_zerolines=show_zerolines,
        equal_scale_axes=equal_scale_axes,
    )


def _scatter_plot_2d_noninteractive(
    df: pd.DataFrame,
    x: Union[str, pd.Series, np.ndarray],
    y: Union[str, pd.Series, np.ndarray],
    color_col: Optional[str] = None,
    figsize: Tuple[float, float] = (12, 12),
    opacity: float = 0.8,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    show_xticks: bool = False,
    show_yticks: bool = False,
    show_zerolines: bool = False,
    equal_scale_axes: bool = False,
) -> None:
    _fig, ax = plt.subplots(figsize=figsize)
    if color_col is not None:
        is_color_categorical = _is_color_col_categorical(df[color_col])
        if is_color_categorical:
            color_categorical = pd.Categorical(df[color_col])
            color_codes = color_categorical.codes
        else:
            color_codes = df[color_col]
        scatter = ax.scatter(x, y, c=color_codes, alpha=opacity)

        if is_color_categorical:
            legend_elements = scatter.legend_elements()
            ax.legend(legend_elements[0], color_categorical.categories, title=color_col)
        else:
            cbar = plt.colorbar(scatter)
            cbar.ax.set_ylabel(color_col)
    else:
        ax.scatter(x, y, alpha=opacity)

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if not show_xticks:
        ax.set_xticks([])
    if not show_yticks:
        ax.set_yticks([])
    if not show_zerolines:
        ax.grid(False)
    if equal_scale_axes:
        ax.set_aspect("equal", "datalim")
    plt.show()


def _scatter_plot_2d_interactive(
    df: pd.DataFrame,
    x: Union[str, pd.Series, np.ndarray],
    y: Union[str, pd.Series, np.ndarray],
    color_col: Optional[str] = None,
    figsize: Tuple[float, float] = (12, 12),
    opacity: float = 0.8,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    show_xticks: bool = False,
    show_yticks: bool = False,
    show_zerolines: bool = False,
    equal_scale_axes: bool = False,
) -> None:
    layout = go.Layout(
        width=figsize[0] * _INCHES_TO_PIXELS,
        height=figsize[1] * _INCHES_TO_PIXELS,
        xaxis=go.layout.XAxis(
            showgrid=False, showticklabels=show_xticks, zeroline=show_zerolines, title=xlabel
        ),
        yaxis=go.layout.YAxis(
            showgrid=False, showticklabels=show_yticks, zeroline=show_zerolines, title=ylabel
        ),
        legend=go.layout.Legend(title=f"<b>{color_col}</b>"),
    )
    if equal_scale_axes:
        layout.yaxis.scaleanchor = "x"
        layout.yaxis.scaleratio = 1
    fig = go.Figure(layout=layout)
    if color_col is not None:
        is_color_categorical = _is_color_col_categorical(df[color_col])
        if is_color_categorical:
            df = df.copy()
            x_name, y_name = "__edvart_scatter_x", "__edvart_scatter_y"
            df[x_name] = x
            df[y_name] = y
            for group_name, group in df.groupby(color_col):
                fig.add_trace(
                    go.Scatter(
                        x=group[x_name],
                        y=group[y_name],
                        mode="markers",
                        marker=dict(opacity=opacity),
                        name=group_name,
                        text=[
                            "</br>".join(
                                f"{col_name}: {df.loc[row, col_name]}"
                                for col_name in group.columns.drop([x_name, y_name])
                            )
                            for row in group.index
                        ],
                    )
                )
        else:
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=y,
                    mode="markers",
                    marker=dict(
                        color=df[color_col], opacity=opacity, colorbar=dict(title=color_col)
                    ),
                    text=[
                        "</br>".join(
                            f"{col_name}: {df.loc[row, col_name]}" for col_name in df.columns
                        )
                        for row in df.index
                    ],
                ),
            )
    else:  # color_col is None
        fig.add_trace(
            go.Scatter(
                x=x,
                y=y,
                mode="markers",
                marker=dict(opacity=opacity),
                text=[
                    "</br>".join(f"{col_name}: {df.loc[row, col_name]}" for col_name in df.columns)
                    for row in df.index
                ],
            ),
        )

    fig.show()



import uuid
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Any, Dict, List, Optional

import pandas as pd


class Verbosity(IntEnum):
    """
    Verbosity of the exported code.
        LOW
            A single function call generates the entire section.
        MEDIUM
            Function calls to parameterizable functions are generated for each column separately
            in separate cells.
        HIGH
            Similar to MEDIUM, but in addition, function definitions are generated, column
            data type inference and default statistics become customizable.
    """

    LOW = 1
    MEDIUM = 2
    HIGH = 3


class Section(ABC):
    """Base class for report sections and subsections.

    Parameters
    -----------
    verbosity : Verbosity
        The verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns that are considered in the analysis of the section.
        All columns are used by default.

    Notes
    -----
    To create a new section, subclass this class and implement the abstract methods.

    * `__init__` initializes your object and accepts `verbosity` and `columns`
      (in addition to any other section specific parameters).
        - `verbosity` is an enum representing the detail level of the exported code.
        - `columns` is a list of names of columns which will be used in the analysis.
    * `required_imports` returns a list of lines of code that import the packages
       required by the analysis which will get added to a cell at the top of the exported notebook.
       Keep in mind that different verbosity levels usually require a different set of imports.
    * `add_cells(cells)` adds cells to the list of cells `cells`.
       This method is used to build the code for the exported notebook.
        - To create a markdown cell, pass a string to `nbformat.v4.new_markdown_cell()`
        - To create a code cell pass a string to `nbformat.v4.new_code_cell()`
        - Finally append the objects returned by the functions mentioned above to `cells`
        - Keep in mind that the code created should conform to `verbosity`
    * `show` renders the analysis in place in the calling notebook.
    """

    def __init__(self, verbosity: Verbosity = Verbosity.LOW, columns: Optional[List[str]] = None):
        self.verbosity = Verbosity(verbosity)
        self.columns = columns
        self._section_id: str = str(uuid.uuid4())

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the section.

        Returns
        -------
        str
            Name of the section.
        """

    @property
    def uid(self) -> str:
        """Identifier of the section used for generating table of contents.

        Should be unique across sections.

        Returns
        -------
        str
            Unique identifier of the section
        """
        return self._section_id

    def get_title(self, section_level: int) -> str:
        """Gets the title of the section in markdown format.

        Includes a hyperlink id tag that is used by the table of contents.

        Parameters
        ----------
        section_level: int
            The level of the section. Adds # according to it. Highest level sections should have it
            set to 1.

        Returns
        -------
        str
            Title of the section in markdown format.
        """
        title = f"{'#' * section_level} {self.name}<a id='{self.uid}'>"
        if section_level == 1:
            title += "\n---"
        return title

    @abstractmethod
    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np']
        """

    @abstractmethod
    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
            The dictionaries can be generated with nbformat.v4.new_code_cell() and/or
            nbformat.v4.new_markdown_cell().
        """

    @abstractmethod
    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output in the calling notebook using IPython.display.display().

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """


class ReportSection(Section):
    """Base class for top level report sections.

    Contains subsections which are also of subtype Section and implement the report generation.

    Parameters
    ----------
    subsections : List[Section]
        List of subsections that should be contained in this top level section
    verbosity : Verbosity
        The verbosity of the code generated in the exported notebook.
    columns : List[str], optional
        List of columns that are considered in the analysis of the section,
        all columns are used by default
    """

    def __init__(
        self,
        subsections: List[Section],
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
    ):
        super().__init__(verbosity, columns)
        self.subsections = subsections

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ['import pandas as pd', 'import numpy as np'].
        """
        imports_set = set()
        for subsec in self.subsections:
            imports_set.update(subsec.required_imports())

        return list(imports_set)

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        for subsec in self.subsections:
            subsec.add_cells(cells=cells, df=df)

    def show(self, df: pd.DataFrame) -> None:
        """Generates cell output of this section in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output.
        """
        for subsec in self.subsections:
            subsec.show(df)



from datetime import datetime
from itertools import takewhile
from typing import Any, Callable, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import pandas as pd
import seaborn as sns
from IPython.display import Markdown, display

from edvart.data_types import is_numeric
from edvart.decorators import check_index_time_ascending
from edvart.report_sections.code_string_formatting import get_code
from edvart.report_sections.section_base import Section, Verbosity


class BoxplotsOverTime(Section):
    """Generates the boxplots over time intervals section.

    For each column, generates a series of boxes, each box representing distribution of values
    in the given column during a time interval.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the generated code in the exported notebook.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    grouping_function : Callable[[Any], str], optional
        Function to group the data into intervals. Cannot pass an anonymous function, i.e. the
        function must be assigned an identifier.
        To pass a lambda, simply assign it to a variable and pass the variable.
        If None is passed, a default grouping will be selected (see `default_nunique_max`).
    grouping_function_imports: List[str], optional
        Additional imports required for the grouping function.
    grouping_name : str, optional
        Name of grouping, will be displayed as title of the horizontal axis.
    default_nunique_max : int (default = 80)
        If no grouping function is passed, the most granular grouping which produces at most
        `default_nunique_max` unique values is selected from the following:
        Hour, Day, Week, Month, Quarter, Year, Decade.
        If a default grouping is selected,
        a corresponding name is displayed on the horizontal axis by default
    """

    def __init__(
        self,
        verbosity: Verbosity = Verbosity.LOW,
        columns: Optional[List[str]] = None,
        grouping_function: Callable[[Any], str] = None,
        grouping_function_imports: Optional[List[str]] = None,
        grouping_name: Optional[str] = None,
        default_nunique_max: int = 80,
    ):
        self.grouping_function = grouping_function
        self.grouping_function_imports = grouping_function_imports
        self.grouping_name = grouping_name
        self.default_nunique_max = default_nunique_max
        super().__init__(verbosity, columns)

    @property
    def name(self) -> str:
        return "Boxplots over time intervals"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"].
        """
        if self.verbosity <= Verbosity.MEDIUM:
            imports = [
                "from edvart.report_sections.timeseries_analysis.boxplots_over_time"
                " import show_boxplots_over_time"
            ]
        else:
            imports = [
                "from datetime import datetime",
                "import matplotlib.pyplot as plt",
                "import seaborn as sns",
                "from IPython.display import display, Markdown",
                "from edvart.data_types import is_numeric",
            ]
        if self.grouping_function_imports is not None:
            imports.extend(self.grouping_function_imports)
        return imports

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        cells.append(section_header)
        if self.grouping_function is not None:
            grouping_func_code = get_code(self.grouping_function) + "\n\n"
            if grouping_func_code.startswith("def"):
                grouping_func_name = self.grouping_function.__name__
            else:
                # function was created using an assignment to a variable
                # -> get_code returns <var_name> = <lambda or function> -> we want <var_name>
                grouping_func_name = "".join(
                    takewhile(
                        lambda char: char.isalnum() or char == "_",
                        list(grouping_func_code),
                    )
                )
            cells.append(nbfv4.new_code_cell(grouping_func_code))
        default_call = "show_boxplots_over_time(df=df"
        if self.columns is not None:
            default_call += f", columns={self.columns}"
        if self.grouping_function is not None:
            default_call += f", grouping_function={grouping_func_name}"
        if self.grouping_name is not None:
            default_call += f", grouping_name='{self.grouping_name}'"
        default_call += ")"

        if self.verbosity <= Verbosity.MEDIUM:
            code = default_call
        else:
            if self.grouping_function is None:
                code = (
                    get_code(default_grouping_functions)
                    + "\n\n"
                    + get_code(get_default_grouping_func)
                    + "\n\n"
                )
            else:
                code = ""

            code += get_code(show_boxplots_over_time) + "\n\n" + default_call

        cells.append(nbfv4.new_code_cell(code))

    def show(self, df: pd.DataFrame) -> None:
        """Generates boxplots grouped over time intervals in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        show_boxplots_over_time(
            df=df,
            columns=self.columns,
            grouping_function=self.grouping_function,
            grouping_name=self.grouping_name,
            default_nunique_max=self.default_nunique_max,
        )


def default_grouping_functions() -> Dict[str, Callable[[datetime], str]]:
    """Return a dictionary of function names and functions.

    The function takes a pandas datetime and represents it as a rougher (in terms of time)
    string, which can be used for grouping.
    Available groupings are: Hour, Day, Week, Month, Quarter, Year, Decade

    Returns
    -------
    Dict[str, Callable[[datetime], str]]
        Dictionary from grouping function names to grouping functions.
    """
    return {
        "Hour": lambda x: f"{x.day}/{x.month}/{x.year} {x.hour}:00",
        "Day": lambda x: f"{x.day}/{x.month}/{x.year}",
        "Week": lambda x: f"W{x.week}, {x.year if x.dayofweek < x.dayofyear else x.year - 1}",
        "Month": lambda x: f"{x.month_name()[:3]} {x.year}",
        "Quarter": lambda x: f"Q{x.quarter} {x.year}",
        "Year": lambda x: f"{x.year}",
        "Decade": lambda x: f"{x.year // 10 * 10}s",
    }


def get_default_grouping_func(df: pd.DataFrame, nunique_max: int = 80) -> Tuple[str, Callable]:
    """Return the most granular function to group df.index into at most nunique_max intervals.

    Uses grouping functions from `default_grouping_functions`.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe, for index of which a suitable grouping
    nunique_max : int (default = 80)
        Maximum number of intervals to group to.

    Returns
    -------
    Tuple[str, Callable]
        Name of selected grouping function and the grouping function itself.
    """
    # find most granular grouping which does not produce too many values
    index_series = df.index.to_series()
    items = list(default_grouping_functions().items())
    assert items  # check that there are some values

    for name, func in items:
        if index_series.apply(func).nunique() <= nunique_max:
            return name, func
    # If no grouping is rough enough, use the roughest available
    return items[-1]


@check_index_time_ascending
def show_boxplots_over_time(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    grouping_function: Callable[[Any], str] = None,
    grouping_name: Optional[str] = None,
    default_nunique_max: int = 80,
    figsize: Tuple[float, float] = (20, 7),
    color: Any = None,
) -> None:
    """Generate boxplots over time intervals.

    For each column, generates a series of boxes, each box representing distribution of values
    in the given column during a time interval.

    Parameters
    ----------
    df : pd.DataFrame
        Data to analyze.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    grouping_function : Callable[[Any], str], optional
        Function to group the data into intervals. Cannot pass an anonymous function, i.e. the
        function must be assigned an identifier.
        To pass a lambda, simply assign it to a variable and pass the variable.
        If None is passed, a default grouping will be selected (see `default_nunique_max`).
    grouping_name : str, optional
        Name of grouping, will be displayed as title of the horizontal axis.
    default_nunique_max : int, optional
        If no grouping function is passed, the most granular grouping which produces at most
        `default_nunique_max` unique values is selected from the following:
        Hour, Day, Week, Month, Quarter, Year, Decade.
        If a default grouping is selected,
        a corresponding name is displayed on the horizontal axis by default
    figsize : Tuple[float, float] (default = (20, 7))
        Size of boxplot series figure for each column.
    color : Any, optional
        Color or color map compatible with matplotlib/seaborn.
        By default a "rainbow" color map - color of individual boxes changes over time.

    Raises
    ------
    ValueError
        If the input data is not indexed by time in ascending order.
    """
    default_grouping_funcs = default_grouping_functions()
    if grouping_function is None:
        grouping_name, grouping_function = get_default_grouping_func(
            df, nunique_max=default_nunique_max
        )
    elif default_grouping_funcs.get(grouping_name) is not None:
        grouping_function = default_grouping_funcs[grouping_name]

    if columns is None:
        columns = [col for col in df.columns if is_numeric(df[col])]

    for column in columns:
        if not is_numeric(df[column]):
            raise ValueError(f"Cannot plot boxplot for non-numeric column `{column}`")
        display(Markdown(f"---\n### {column}"))
        ax = sns.boxplot(
            x=df.index.to_series().apply(grouping_function),
            y=df[column],
            color=color,
        )

        ax.set_xticklabels(ax.get_xticklabels(), rotation=45)
        if grouping_name is not None:
            ax.set_xlabel(grouping_name)
        ax.figure.set_size_inches(*figsize)
        plt.show()



import os

import pandas as pd

_DATASETS_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "../example-datasets"))


def dataset_auto() -> pd.DataFrame:
    """
    Returns a dataset containing information on the technical specifications of cars.
    The dataset contains 398 rows and 9 columns. There are no missing values.

    Source: https://www.kaggle.com/datasets/uciml/autompg-dataset
    Unmodified.

    Returns
    -------
    pd.DataFrame
    """
    return pd.read_csv(os.path.join(_DATASETS_DIR, "auto.csv"))


def dataset_meteorite_landings() -> pd.DataFrame:
    """
    Returns a dataset that includes the location, mass, composition, and fall year
    for over 45,000 meteorites that have struck our planet.

    Source of data: NASA Open Data Portal
    (https://data.nasa.gov/Space-Science/Meteorite-Landings/gh4g-9sfh)

    Returns
    -------
    pd.DataFrame
    """
    return pd.read_csv(os.path.join(_DATASETS_DIR, "Meteorite_Landings.csv"))


def dataset_titanic() -> pd.DataFrame:
    """
    Returns a dataset that contains data for 891 of the real Titanic passengers.
    Each row represents one person. The columns describe different attributes about the person
    including whether they survived, their age, their passenger-class, their sex
    and the fare they paid.

    The dataset contains 891 rows and 12 columns.


    Source: https://www.kaggle.com/datasets/hesh97/titanicdataset-traincsv
    Unmodified.

    Returns
    -------
    pd.DataFrame
    """
    return pd.read_csv(os.path.join(_DATASETS_DIR, "titanic.csv"))


def dataset_global_temp() -> pd.DataFrame:
    """
    Returns a time-series dataset containing monthly deviations from mean global average temperature
    from 1880 until 2016 according to two methodologies: GCAG and GISTEMP.

    Source: https://datahub.io/core/global-temp.
    Modified: Moved each methodology into its own column.
    """
    return pd.read_csv(
        os.path.join(_DATASETS_DIR, "global_temp.csv"),
        index_col="Date",
        parse_dates=["Date"],
    )


def dataset_pollution() -> pd.DataFrame:
    """
    Returns a time-series dataset containing hourly weather and pollution data from 2010 until 2014
    from Beijing, China. There are 43800 rows and 8 columns.

    Source: https://www.kaggle.com/datasets/djhavera/beijing-pm25-data-data-set
    Modified:
    - Merged columns "year", "month", "day", "hour" into a single "date" column.
    - Removed the first day, for which pollution data is missing.
    - Renamed columns.
    """
    return pd.read_csv(
        os.path.join(_DATASETS_DIR, "pollution.csv"),
        index_col=["date"],
        parse_dates=["date"],
    )



from typing import Any, Dict, List

import nbformat.v4 as nbfv4
from IPython.display import Markdown, display

from edvart.report_sections.section_base import ReportSection, Section


class TableOfContents:
    """Generates the Table of Contents section of the report.

    Parameters
    ----------
    include_subsections: bool
        A boolean controlling whether the subsections should be included in the table of contents.
        However, they won't be included in an exported notebook created by report's
        export_notebook function.
    """

    def __init__(self, include_subsections: bool):
        self._include_subsections = include_subsections

    @property
    def _title(self) -> str:
        return "# Table of Contents\n---"

    @staticmethod
    def _get_section_link(section: Section, section_level: int) -> str:
        """Gets the section link in the markdown format.

        Parameters
        ----------
        section: Section
            Section for which the link should be generated.

        section_level: int
            The level of the section. Highest level sections should have it set to 1.

        Returns
        -------
        str
            The section link in the markdown format.
        """
        beginning_whitespace = "\t" * (section_level - 1)
        return f"{beginning_whitespace} * [{section.name}](#{section.uid})\n"

    def _add_section_lines(
        self, section: Section, section_level: int, lines: List[str], include_subsections: bool
    ) -> None:
        """Generates table of contents' section's cell output in the calling notebook.

        Parameters
        ----------
        section: Section
            Section for which table of contents' lines are added to the provided list.
        section_level: int
            The level of the section. Highest level sections should have it set to 1.
        lines: List[str]
            Lines that should be added to the table of content.
        include_subsections: bool
            A boolean controlling whether the subsections' lines should be added as well.
        """
        lines.append(TableOfContents._get_section_link(section, section_level))

        if isinstance(section, ReportSection) and include_subsections:
            for subsection in section.subsections:
                self._add_section_lines(subsection, section_level + 1, lines, True)

    def add_cells(self, sections: List[Section], cells: List[Dict[str, Any]]) -> None:
        """Adds table of contents cells to the list of cells. The subsections won't be included.

        Parameters
        ----------
        sections: List[Section]
            List of sections that should be included in the table of contents.
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries.
        """
        section_header = nbfv4.new_markdown_cell(self._title)
        cells.append(section_header)

        lines: List[str] = []
        for section in sections:
            lines.append(TableOfContents._get_section_link(section, 1))
        cells.append(nbfv4.new_markdown_cell("\n".join(lines)))

    def show(self, sections: List[Section]) -> None:
        """Generates table of contents' cell output in the calling notebook.

        Parameters
        ----------
        sections: List[Section]
            List of sections that should be included in the table of contents.
        """
        display(Markdown(self._title))

        lines = []
        for section in sections:
            self._add_section_lines(section, 1, lines, self._include_subsections)
        display(Markdown("\n".join(lines)))



import functools
from typing import Any, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import nbformat.v4 as nbfv4
import numpy as np
import pandas as pd
from IPython.display import Markdown, display
from statsmodels.graphics import tsaplots

from edvart.data_types import is_numeric
from edvart.decorators import check_index_time_ascending
from edvart.report_sections.code_string_formatting import get_code
from edvart.report_sections.section_base import Section, Verbosity


class Autocorrelation(Section):
    """Generates autocorrelation (ACF) and partial autocorrelation function (PACF) subsection.

    Parameters
    ----------
    verbosity : Verbosity (default = Verbosity.LOW)
        Verbosity of the generated code in the exported notebook.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    """

    @property
    def name(self) -> str:
        return "Autocorrelation"

    def required_imports(self) -> List[str]:
        """Returns a list of imports to be put at the top of a generated notebook.

        Returns
        -------
        List[str]
            List of import strings to be added at the top of the generated notebook,
            e.g. ["import pandas as pd", "import numpy as np"].
        """
        if self.verbosity == Verbosity.LOW:
            return [
                "from edvart.report_sections.timeseries_analysis.autocorrelation"
                " import show_autocorrelation"
            ]
        if self.verbosity == Verbosity.MEDIUM:
            return [
                "from edvart.report_sections.timeseries_analysis.autocorrelation"
                " import plot_acf, plot_pacf"
            ]
        # Verbosity.HIGH
        return [
            "import functools",
            "import matplotlib.pyplot as plt",
            "import statsmodels.graphics.tsaplots as tsaplots",
            "from IPython.display import display, Markdown",
            "from edvart.data_types import is_numeric",
        ]

    def add_cells(self, cells: List[Dict[str, Any]], df: pd.DataFrame) -> None:
        """Adds cells to the list of cells.

        Cells can be either code cells or markdown cells.

        Parameters
        ----------
        cells : List[Dict[str, Any]]
            List of generated notebook cells which are represented as dictionaries
        df: pd.DataFrame
            Data for which to add the cells.
        """
        if self.verbosity == Verbosity.LOW:
            section_header = nbfv4.new_markdown_cell(self.get_title(section_level=2))
            default_call = "show_autocorrelation(df=df"
            if self.columns is not None:
                default_call += f", columns={self.columns}"
            default_call += ")"
            cells.append(section_header)
            cells.append(nbfv4.new_code_cell(default_call))
            return

        section_header_acf = nbfv4.new_markdown_cell(self.get_title(section_level=2))
        default_call_acf = "plot_acf(df=df"
        if self.columns is not None:
            default_call_acf += f", columns={self.columns}"
        default_call_acf += ")"

        section_header_pacf = nbfv4.new_markdown_cell("## Partial autocorrelation")
        default_call_pacf = "plot_pacf(df=df"
        if self.columns is not None:
            default_call_pacf += f", columns={self.columns}"
        default_call_pacf += ")"

        if self.verbosity == Verbosity.MEDIUM:
            code_acf = default_call_acf
            code_pacf = default_call_pacf
        else:
            code_acf = get_code(plot_acf) + "\n\n" + default_call_acf
            code_pacf = get_code(plot_pacf) + "\n\n" + default_call_pacf

        cells.append(section_header_acf)
        cells.append(nbfv4.new_code_cell(code_acf))
        cells.append(section_header_pacf)
        cells.append(nbfv4.new_code_cell(code_pacf))

    def show(self, df: pd.DataFrame) -> None:
        """Generates ACF and PACF plots in the calling notebook.

        Parameters
        ----------
        df : pd.DataFrame
            Data based on which to generate the cell output
        """
        display(Markdown(self.get_title(section_level=2)))
        show_autocorrelation(df=df, columns=self.columns)


@check_index_time_ascending
def show_autocorrelation(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    lags: Optional[List[int]] = None,
    figsize: Tuple[float, float] = (15, 6),
) -> None:
    """Generate autocorrelation (ACF) and partial autocorrelation function (PACF) plots.

    ACF returns, for a given lag, correlation between
    a given timeseries and itself shifted by this lag.

    Partial autocorrelation function returns conditional autocorrelation given all smaller
    lag values up to the given lag.

    Parameters
    ----------
    df: pd.DataFrame
        Data to analyze.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    lags: List[int], optional
        List of lag values to plot ACF for
    figsize : Tuple[float, float] (default = (15, 6))
        Size of figure of (partial) autocorrelation plot.

    Raises
    ------
    ValueError
        If the input data is not indexed by time in ascending order.
    """
    plot_acf(df, columns, lags, figsize, partial=False)
    display(Markdown("## Partial autocorrelation"))
    plot_pacf(df, columns, lags, figsize)


@check_index_time_ascending
def plot_acf(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    lags: Optional[List[int]] = None,
    figsize: Tuple[float, float] = (15, 6),
    partial: bool = False,
) -> None:
    """Plot ACF or PACF.

    Autocorrelation function (ACF) returns, for a given lag, correlation between
    the timeseries and itself shifted by this lag.

    Partial autocorrelation function (PACF) returns conditional autocorrelation given all
    smaller lag values up to the given lag.

    Parameters
    ----------
    df: pd.DataFrame
        Data to analyze.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    lags: List[int], optional
        List of lag values to plot ACF for
    figsize : Tuple[float, float] (default = (15, 6))
        Size of figure of (partial) autocorrelation plot.
    partial : bool (default = False)
        If True, PACF will be plotted, otherwise, ACF will be plotted.
    """
    if columns is None:
        columns = [col for col in df.columns if is_numeric(df[col])]
    else:
        for col in columns:
            if not is_numeric(df[col]):
                raise ValueError(f"Cannot plot autocorrelation for non-numeric column `{col}`")
    plot_func = (
        functools.partial(tsaplots.plot_pacf, method="ywm") if partial else tsaplots.plot_acf
    )
    if partial and lags is None:
        nobs = len(df)
        # See https://github.com/statsmodels/statsmodels/blob/01b19d7d111b29c183f620ff0a949ef6391ff8ee/statsmodels/graphics/tsaplots.py#L20
        default_lags_limit = int(np.ceil(10 * np.log10(nobs))) + 1
        # Partial function can only be computed for lags up to nobs // 2 - 1
        lags_limit = min(default_lags_limit, nobs // 2)
        lags = list(range(lags_limit))
    for col in columns:
        display(Markdown(f"---\n### {col}"))
        fig = plot_func(df[col].dropna(), lags=lags)
        ax = fig.axes[0]
        ax.set_title("")
        ax.set_xlabel("Lag")
        ax.set_ylabel(("Partial " if partial else "") + "Autocorrelation")
        fig.set_size_inches(*figsize)
        plt.show()


@check_index_time_ascending
def plot_pacf(
    df: pd.DataFrame,
    columns: Optional[List[str]] = None,
    lags: Optional[List[int]] = None,
    figsize: Tuple[float, float] = (15, 6),
) -> None:
    """Plot PACF.

    Partial autocorrelation function (PACF) returns conditional autocorrelation given all
    smaller lag values up to the given lag.

    Parameters
    ----------
    df: pd.DataFrame
        Data to analyze.
    columns : List[str], optional
        List of columns to analyze. Only numeric column can be analyzed.
        All numeric columns are analyzed by default.
    lags: List[int], optional
        List of lag values to plot ACF for
    figsize : Tuple[float, float] (default = (15, 6))
        Size of figure of (partial) autocorrelation plot.
    """
    plot_acf(df, columns, lags, figsize, partial=True)



from typing import Any, Dict, List, Union

import pandas as pd
from IPython.display import display
from pandas.io.formats.style import Styler


def series_to_frame(series: pd.Series, index_name: str, column_name: str) -> pd.DataFrame:
    """Converts a pandas.Series to a pandas.DataFrame by putting the series index into a separate
    column.

    Parameters
    ---
    series : pd.Series
        Input series
    index_name : str
        Name of the new column into which the series index will be put
    column_name : str
        Name of the series values column

    Returns
    ---
    pd.DataFrame
        Dataframe with two columns index_name and column_name with values of series.index and
        series.values respectively
    """
    return series.rename_axis(index=index_name).to_frame(name=column_name).reset_index()


def render_dictionary(dictionary: Dict[str, Any]) -> None:
    """
    Converts a dictionary to a dataframe and renders that dataframe in the report notebook.

    Parameters
    ---
    dictionary: Dict['str', Any]
        Dictionary to be rendered
    """
    dictionary = {key: str(value) for key, value in dictionary.items()}

    display(pd.DataFrame.from_dict(dictionary, orient="index", columns=[""]))


def dict_to_html(dictionary: Dict[str, Any]) -> str:
    """Converts a dictionary to a dataframe in HTML string form.

    Parameters
    ----------
    dictionary : Dict['str', Any]
        DictDictionary to be converted
    """
    dictionary = {key: str(value) for key, value in dictionary.items()}

    return pd.DataFrame.from_dict(dictionary, orient="index", columns=[""]).to_html()


def add_html_heading(html: str, heading: str, heading_level: int = 2) -> str:
    """
    Adds a heading to an HTML string with the specified text and heading level.

    Parameters
    ---
    html : str
        HTML string to which to add heading
    heading : str
        Text of the heading
    heading_level : int
        Level of the heading

    Returns
    ---
    str
        HTML string with heading added
    """
    return f"<h{heading_level}> {heading} </h{heading_level}>" + html


def subcells_html(elements: List[List[str]]) -> str:
    """
    Returns HTML table in string format according to the elements matrix.

    Parameters
    ---
    elements : List[List[str]]
        Elements which should be rendered in table cells, outer list represents rows,
        inner list represents columns, elements themselves should be HTML strings

    Returns
    ---
    str
        Table in HTML string ready to be rendered for example by IPython.display.display_html
    """
    default_cell_css = """
        "
        text-align: left;
        vertical-align: text-top;
        align-content: center;
        "
    """

    i = 0
    html = '<table width="100%">'
    for row in elements:
        html += "<tr>"
        for element in row:
            if i == 0:
                html += f"<td style={default_cell_css}> {element} </td>"
                i = 1
            else:
                html += f"<td style={default_cell_css}> {element} </td>"
                i = 0
        html += "</tr>"
    html += "</table>"

    return html


def format_number(
    number: Union[int, float], decimal_places: int = 2, thousand_separator: str = ""
) -> str:
    """
    Formats a number by truncating decimal places (if it is a float) and optionally adds thousand
    separators.

    Parameters
    ---
    number : Union[int, float]
        Number to be converted to string
    decimal_places : int
        Number of decimal places in case of float
    thousand_separator : str
        Character or string with which thousands should be separated

    Returns
    ---
    str
        Formatted number in a string representation
    """
    if float(number).is_integer():
        formatted = f"{number:,}"
    else:
        formatted = f"{number:,.0{decimal_places}f}"
    return formatted.replace(",", thousand_separator)


def hide_index(df: pd.DataFrame) -> Styler:
    """
    Hides the index of a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame where the index should be hidden.

    Returns
    -------
    Styler
        Styler object with the index hidden.
    """
    return df.style.hide(axis="index")



import base64
import inspect
import os


def embed_image_base64(image_path: str, mime: str = "image/png") -> str:
    """
    Loads content of an image and embeds it as base64 data URL into the template.
    Intended to be used as a Jinja filter.

    Example Jinja filter usage (CSS):
    ```
    #notebook-container {
            background-image: url('{{ 'background.png' | embed_image_base64('image/png') }}');
    ```

    Parameters
    ----------
    image_path : str
        Relative path from the current template to the image.
    mime : str
        Mime type of the image.

    Returns
    -------
    str
        Data URL with embedded image and mime type specified.
    """
    # Look up directory where currently executed template is located
    # Jinja's @environmentfilter or @contextfilter does not seem to provide
    # any information about the path of the template.
    template_dir = os.path.dirname(inspect.getfile(inspect.currentframe().f_back))
    with open(os.path.join(template_dir, image_path), "rb") as img:
        return f"data:{mime};base64," + str(base64.b64encode(img.read()).decode("utf-8"))
